//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Application\CDirectX11.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CDirectX11.h"

namespace App
{
	CDirectX11::CDirectX11()
	{
		m_pWinData							= NULL;
		m_pDX11Data							= new DX11DATA();
	}

	CDirectX11::CDirectX11(const CDirectX11& other)
	{
	}

	CDirectX11::~CDirectX11()
	{
	}

	// Method for handling display data.
	int CDirectX11::HandleDisplayData(u32* numeratorOut, u32* denominatorOut)
	{
		int retCode = Util::Err::NONE;

		IDXGIFactory*						factory;
		IDXGIAdapter*						adapter;
		IDXGIOutput*						adapterOutput;
		DXGI_MODE_DESC*						displayModeList;
		u32									numModes;
		
		if(FAILED(CreateDXGIFactory(__uuidof(IDXGIFactory), (void**)&factory)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_FACTORY;
		}

		// Retrieve adapter, and display information.
		if(FAILED(factory->EnumAdapters(0, &adapter)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_RETRIEVE_PRIMARY_ADAPTER;
		}

		if(FAILED(adapter->EnumOutputs(0, &adapterOutput)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_RETRIEVE_PRIMARY_MONITOR;
		}

		if(FAILED(adapterOutput->GetDisplayModeList(DXGI_FORMAT_R8G8B8A8_UNORM, DXGI_ENUM_MODES_INTERLACED, &numModes, NULL)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_RETRIEVE_MONITOR_DISPLAY_MODE_COUNT;
		}

		displayModeList = new DXGI_MODE_DESC[numModes];
		
		if(FAILED(adapterOutput->GetDisplayModeList(DXGI_FORMAT_R8G8B8A8_UNORM, DXGI_ENUM_MODES_INTERLACED, &numModes, displayModeList)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_RETRIEVE_MONITOR_DISPLAY_MODE_LIST;
		}

		// Find a monitor that can support this application.
		for(u32 i = 0; i < numModes; i++)
		{
			if(displayModeList[i].Width <= (u32)(m_pWinData->winWidth))
			{
				if(displayModeList[i].Height <= (u32)(m_pWinData->winHeight))
				{
					*numeratorOut = displayModeList[i].RefreshRate.Numerator;
					*denominatorOut = displayModeList[i].RefreshRate.Denominator;
					break;
				}
			}
		}

		// Store the adapter description.
		if(FAILED(adapter->GetDesc(&m_pDX11Data->adapterDesc)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_RETRIVE_ADAPTER_DESCRIPTION;
		}

		m_pDX11Data->videoCardMemory = m_pDX11Data->adapterDesc.DedicatedVideoMemory / BYTES_TO_MBS;
		wcstombs_s(NULL, m_pDX11Data->videoCardDesc, 128, m_pDX11Data->adapterDesc.Description, 128);

		SAFE_DELETE_ARRAY(displayModeList);
		SAFE_RELEASE(adapterOutput);
		SAFE_RELEASE(adapter);
		SAFE_RELEASE(factory);

		return retCode;
	}

	// Method for creating the swap chain, device, and device context.
	int CDirectX11::CreateSwapChainAndDevices(u32 numerator, u32 denominator)
	{
		int retCode = Util::Err::NONE;

		DXGI_SWAP_CHAIN_DESC				swapChainDesc;
		D3D_FEATURE_LEVEL					featureLevel;
		
		// Setup the swap chain description.
		ZeroMemory(&swapChainDesc, sizeof(swapChainDesc));

		swapChainDesc.BufferCount							= 1;
		swapChainDesc.BufferDesc.Width						= m_pWinData->width;
		swapChainDesc.BufferDesc.Height						= m_pWinData->height;
		swapChainDesc.BufferDesc.Format						= DXGI_FORMAT_R8G8B8A8_UNORM;

		swapChainDesc.BufferDesc.RefreshRate.Numerator		= m_pWinData->bVSync ? numerator : 0;
		swapChainDesc.BufferDesc.RefreshRate.Denominator	= m_pWinData->bVSync ? denominator : 1;

		swapChainDesc.BufferUsage							= DXGI_USAGE_RENDER_TARGET_OUTPUT;
		swapChainDesc.OutputWindow							= m_pWinData->hWnd;

		swapChainDesc.SampleDesc							= m_pDX11Data->samplerDesc;

		swapChainDesc.Windowed								= !m_pWinData->bFullscreen;
		
		swapChainDesc.BufferDesc.ScanlineOrdering			= DXGI_MODE_SCANLINE_ORDER_UNSPECIFIED;
		swapChainDesc.BufferDesc.Scaling					= DXGI_MODE_SCALING_STRETCHED;

		swapChainDesc.SwapEffect							= DXGI_SWAP_EFFECT_DISCARD;

		swapChainDesc.Flags									= DXGI_SWAP_CHAIN_FLAG_ALLOW_MODE_SWITCH;

		featureLevel										= D3D_FEATURE_LEVEL_11_0;

		// Create the swapchain, device, and device context.
		if(FAILED(D3D11CreateDeviceAndSwapChain(NULL, D3D_DRIVER_TYPE_HARDWARE, NULL, 0, &featureLevel, 1, D3D11_SDK_VERSION, &swapChainDesc, 
			&m_pDX11Data->pSwapChain, &m_pDX11Data->pDev, NULL, &m_pDX11Data->pDevCon)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_DEVICE_AND_SWAPCHAIN;
		}

		return retCode;
	}

	// Method for creating the render, and depth stencil buffers, views, and states
	int CDirectX11::CreateRenderAndDepthBuffers()
	{
		int retCode = Util::Err::NONE;
		
		ID3D11Texture2D*					backBufferPtr;
		D3D11_TEXTURE2D_DESC				depthBufferDesc;
		D3D11_DEPTH_STENCIL_DESC			depthStencelDesc;
		D3D11_DEPTH_STENCIL_VIEW_DESC		depthStencilViewDesc;
		D3D11_SHADER_RESOURCE_VIEW_DESC		depthStencilResourceViewDesc;
		
		// Get the back buffer and create the main render target view.
		if(FAILED(m_pDX11Data->pSwapChain->GetBuffer(0, __uuidof(ID3D11Texture2D), (void**)&backBufferPtr)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_RETRIEVE_BACK_BUFFEER_POINTER;
		}

		if(FAILED(m_pDX11Data->pDev->CreateRenderTargetView(backBufferPtr, NULL, &m_pDX11Data->pRenderTargetView)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_RENDER_TARGET_VIEW;
		}

		SAFE_RELEASE(backBufferPtr);

		// Setup the depth buffer 
		ZeroMemory(&depthBufferDesc, sizeof(depthBufferDesc));

		depthBufferDesc.Width				= m_pWinData->width;
		depthBufferDesc.Height				= m_pWinData->height;
		depthBufferDesc.MipLevels			= 1;
		depthBufferDesc.ArraySize			= 1;
		depthBufferDesc.Format				= DXGI_FORMAT_R24G8_TYPELESS;
		depthBufferDesc.SampleDesc.Count	= 1;
		depthBufferDesc.SampleDesc.Quality	= 0;
		depthBufferDesc.Usage				= D3D11_USAGE_DEFAULT;
		depthBufferDesc.BindFlags			= D3D11_BIND_DEPTH_STENCIL | D3D11_BIND_SHADER_RESOURCE;
		depthBufferDesc.CPUAccessFlags		= 0;
		depthBufferDesc.MiscFlags			= 0;

		// Create the depth stencil buffer.
		if(FAILED(m_pDX11Data->pDev->CreateTexture2D(&depthBufferDesc, NULL, &m_pDX11Data->pDepthStencilBuffer)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_DEPTH_STENCIL_BUFFER;
		}

		ZeroMemory(&depthStencelDesc, sizeof(depthStencelDesc));

		depthStencelDesc.DepthEnable					= true;
		depthStencelDesc.DepthWriteMask					= D3D11_DEPTH_WRITE_MASK_ALL;
		depthStencelDesc.DepthFunc						= D3D11_COMPARISON_LESS;

		depthStencelDesc.StencilEnable					= false;
		depthStencelDesc.StencilWriteMask				= 0XFF;
		depthStencelDesc.StencilReadMask				= 0XFF;

		depthStencelDesc.FrontFace.StencilFailOp		= D3D11_STENCIL_OP_KEEP;
		depthStencelDesc.FrontFace.StencilDepthFailOp	= D3D11_STENCIL_OP_INCR;
		depthStencelDesc.FrontFace.StencilPassOp		= D3D11_STENCIL_OP_KEEP;
		depthStencelDesc.FrontFace.StencilFunc			= D3D11_COMPARISON_ALWAYS;

		depthStencelDesc.BackFace.StencilFailOp			= D3D11_STENCIL_OP_KEEP;
		depthStencelDesc.BackFace.StencilDepthFailOp	= D3D11_STENCIL_OP_DECR;
		depthStencelDesc.BackFace.StencilPassOp			= D3D11_STENCIL_OP_KEEP;
		depthStencelDesc.BackFace.StencilFunc			= D3D11_COMPARISON_ALWAYS;

		// Create the depth stencil state.
		if(FAILED(m_pDX11Data->pDev->CreateDepthStencilState(&depthStencelDesc, &m_pDX11Data->pDepthStencilState)))
		{
			return Util::Err::ERR_DX11_CREATE_DEPTH_STENCIL_STATE;
		}

		m_pDX11Data->pDevCon->OMSetDepthStencilState(m_pDX11Data->pDepthStencilState, 1);
		ZeroMemory(&depthStencilViewDesc, sizeof(depthStencilViewDesc));

		depthStencilViewDesc.Format					= DXGI_FORMAT_D24_UNORM_S8_UINT;
		depthStencilViewDesc.ViewDimension			= D3D11_DSV_DIMENSION_TEXTURE2D;
		depthStencilViewDesc.Texture2D.MipSlice		= 0;

		// Create the depth stencil view.
		if(FAILED(m_pDX11Data->pDev->CreateDepthStencilView(m_pDX11Data->pDepthStencilBuffer, &depthStencilViewDesc, &m_pDX11Data->pDepthStencilView)))
		{
			return Util::Err::ERR_DX11_FAILED_TO_CREATE_DEPTH_STENCIL_VIEW;
		}

		// Depth stencil resource view.
		ZeroMemory(&depthStencilResourceViewDesc, sizeof(depthStencilResourceViewDesc));

		// Setup the description of the shader resource view.
		depthStencilResourceViewDesc.Format							= DXGI_FORMAT_R24_UNORM_X8_TYPELESS;
		depthStencilResourceViewDesc.ViewDimension					= D3D11_SRV_DIMENSION_TEXTURE2D;
		depthStencilResourceViewDesc.Texture2D.MostDetailedMip		= 0;
		depthStencilResourceViewDesc.Texture2D.MipLevels			= 1;

		// Create the shader resource view.
		if(FAILED(m_pDX11Data->pDev->CreateShaderResourceView(m_pDX11Data->pDepthStencilBuffer, &depthStencilResourceViewDesc, &m_pDX11Data->pDepthStencilResView)))
		{
			return Util::Err::ERR_DX11_CREATE_DEPTH_STENCIL_RESOURCE_VIEW;
		}

		m_pDX11Data->pDevCon->OMSetRenderTargets(1, &m_pDX11Data->pRenderTargetView, m_pDX11Data->pDepthStencilView);
		
		// Create the viewport.
		m_pDX11Data->viewport.Width			= (float)m_pWinData->width;
		m_pDX11Data->viewport.Height		= (float)m_pWinData->height;
		m_pDX11Data->viewport.MinDepth		= 0.0f;
		m_pDX11Data->viewport.MaxDepth		= 1.0f;
		m_pDX11Data->viewport.TopLeftX		= 0.0f;
		m_pDX11Data->viewport.TopLeftY		= 0.0f;

		m_pDX11Data->pDevCon->RSSetViewports(1, &m_pDX11Data->viewport);

		return retCode;
	}

	// Method for initializing DirectX.
	int CDirectX11::Initialize(WINDATA* winData)
	{
		int retCode = Util::Err::NONE;
		
		// Store the 'winData' parameter.
		m_pWinData = winData;

		u32 numerator, denominator;
		
		if((retCode = HandleDisplayData(&numerator, &denominator)) != Util::Err::NONE)
			return retCode;
		
		if((retCode = CreateSwapChainAndDevices(numerator, denominator)) != Util::Err::NONE)
			return retCode;

		if((retCode = CreateRenderAndDepthBuffers()) != Util::Err::NONE)
			return retCode;
		
		return retCode;
	}
	
	int CDirectX11::BeginScene()
	{
		int retCode = Util::Err::NONE;

		m_pDX11Data->pDevCon->ClearRenderTargetView(m_pDX11Data->pRenderTargetView, m_pDX11Data->backgroundColor);
		m_pDX11Data->pDevCon->ClearDepthStencilView(m_pDX11Data->pDepthStencilView, D3D11_CLEAR_DEPTH | D3D11_CLEAR_STENCIL, 1.0f, 0);

		return retCode;
	}

	int CDirectX11::EndScene()
	{
		int retCode = Util::Err::NONE;

		m_pDX11Data->pSwapChain->Present(m_pWinData->bVSync ? 1 : 0, 0);

		return retCode;
	}

	// Method for handling the resize of the application window in relation to DirectX.
	int CDirectX11::ResizeWindow()
	{
		int retCode = Util::Err::NONE;

		if(m_pDX11Data->pSwapChain)
		{
			SAFE_RELEASE(m_pDX11Data->pDepthStencilResView);
			SAFE_RELEASE(m_pDX11Data->pDepthStencilView);
			SAFE_RELEASE(m_pDX11Data->pDepthStencilState);
			SAFE_RELEASE(m_pDX11Data->pDepthStencilBuffer);
			SAFE_RELEASE(m_pDX11Data->pRenderTargetView);

			if(FAILED(m_pDX11Data->pSwapChain->ResizeBuffers(0, m_pWinData->width, m_pWinData->height, DXGI_FORMAT_UNKNOWN, DXGI_SWAP_CHAIN_FLAG_ALLOW_MODE_SWITCH)))
			{
				return Util::Err::ERR_DX11_FAILED_TO_RESIZE_SWAPCHAIN;
			}

			if((retCode = CreateRenderAndDepthBuffers()) != Util::Err::NONE)
				return retCode;
		}
		
		return retCode;
	}
	
	// Method for finding the current swapchain fullscreen state.
	int CDirectX11::FindScreenMode()
	{
		int retCode = Util::Err::NONE;

		if(m_pDX11Data->pSwapChain)
			if(FAILED(m_pDX11Data->pSwapChain->GetFullscreenState((BOOL*)&m_pWinData->bFullscreen, NULL)))
				return Util::Err::ERR_DX11_FAILED_TO_UPDATE_WINDOW_MODE;
		
		return retCode;
	}
	
	// Method for updating the swapchain fullscreen state.
	int CDirectX11::UpdateScreenMode()
	{
		int retCode = Util::Err::NONE;

		if(m_pWinData && (retCode = UpdateScreenMode(m_pWinData->bFullscreen)) != Util::Err::NONE)
			return retCode;

		return retCode;
	}
	
	// Method for updating the swapchain fullscreen state w/ specified fullscreen value.
	int CDirectX11::UpdateScreenMode(bool fullscreen)
	{
		int retCode = Util::Err::NONE;

		if(m_pDX11Data->pSwapChain)
			if(FAILED(m_pDX11Data->pSwapChain->SetFullscreenState(fullscreen, NULL)))
				return Util::Err::ERR_DX11_FAILED_TO_UPDATE_WINDOW_MODE;

		return retCode;
	}

	void CDirectX11::Release()
	{
		SAFE_RELEASE_DELETE(m_pDX11Data);
	}
};