//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CXAudio.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CXAUDIO_H
#define CXAUDIO_H

#include <Windows.h>
#include <XAudio2.h>
#include <x3daudio.h>
#include <string>

#include "..\Globals\CGlobals.h"
#include "..\Utilities\CDebugError.h"
#include "..\Utilities\CMemoryFree.h"
#include "..\Utilities\CDataStack.h"
#include "..\Math\CMathTransform.h"
#include "CWaveFile.h"

#pragma comment(lib, "Xaudio2.lib")

namespace Audio
{
	//---------------------------------------------------------------------------------------------
	// XAudio2Engine
	//---------------------------------------------------------------------------------------------

	// XAudio2Engine class.
	// Class for creating a XAudio2 interface.
	class CXAudio2Engine
	{
	public:
		friend class CXAudio2Source;

	public:
		CXAudio2Engine();
		CXAudio2Engine(const CXAudio2Engine& other);
		virtual ~CXAudio2Engine();

		int									Initialize();
		int									Update();

		void								Release();

		inline void							ListenerTransform(Math::CTransform* pTransform);
		inline Math::CTransform*			ListenerTransform() const;

	private:
		float								m_volume;

		X3DAUDIO_HANDLE						m_X3dInstance;
		X3DAUDIO_LISTENER					m_X3dListener;

		IXAudio2*							m_pXAudio2;
		IXAudio2MasteringVoice*				m_pMasterVoice;

		X3DAUDIO_HANDLE*					m_pX3dInstance;
		X3DAUDIO_LISTENER*					m_pX3dListener;

		Math::CTransform*					m_pListenerTransform;
	};

	// Listener transform getter and setter.

	inline void CXAudio2Engine::ListenerTransform(Math::CTransform* pTransform)
	{
		m_pListenerTransform = pTransform;
	}

	inline Math::CTransform* CXAudio2Engine::ListenerTransform() const
	{
		return m_pListenerTransform;
	}
	
	//---------------------------------------------------------------------------------------------
	// XAudio2Source
	//---------------------------------------------------------------------------------------------

	// CXAudio2Source class.
	// Class used to create a new XAudio2 source. Handles 2d and 3d sounds.
	class CXAudio2Source
	{
	public:
		// Structure for storing data about a source for. Assists in helping with playing multiple 
		// 3d sounds from the same source in different positions.
		struct AUDIO_VOICE
		{
			Math::CVector3			position;
			Math::CVector3			front;
			IXAudio2SourceVoice*	voice;

			AUDIO_VOICE()
			{
				voice = NULL;
			}

			AUDIO_VOICE(const Math::CVector3& Position, const Math::CVector3& Front)
			{
				voice				= NULL;
				position			= Position;
				front				= Front;
			}
		};
	public:
		CXAudio2Source();
		CXAudio2Source(const CXAudio2Source& other);
		virtual ~CXAudio2Source();
		
		int									Initialize(const std::string& filename, CXAudio2Engine*& audioEngine);
		
		int									Update();

		int									Play(float volume = 1.0f, bool loop = false, Math::CVector3 position = Math::CVector3::Zero(), Math::CVector3 front = Math::CVector3::Forward(), AUDIO_VOICE** voiceOut = NULL);
		int									Play(IXAudio2SourceVoice* voice);

		void								Release();

		inline bool							IsPlaying();

		inline bool							Sound3D() const;
		inline void							Sound3D(bool _3dSound);
		
		inline float						DistanceScale() const;
		inline void							DistanceScale(float distanceScale);

		inline float						InnerConeAngle() const;
		inline void							InnerConeAngle(float innerCone);

		inline float						OuterConeAngle() const;
		inline void							OuterConeAngle(float outerCone);

	private:
		std::string							m_filename;

		bool								m_b3dSound;
		bool								m_bLoop;
		float								m_volume;

		float								m_distanceScale;
		X3DAUDIO_EMITTER					m_emitter;
		X3DAUDIO_DSP_SETTINGS				m_dspSettings;
		X3DAUDIO_CONE						m_cone;
		float								m_outerCone;
		float								m_innerCone;
		u32									m_matrixSize;
		float*								m_pMatrix;

		WAVEFORMATEXTENSIBLE				m_wfx;
		XAUDIO2_BUFFER						m_buffer;

		CXAudio2Engine*						m_pXAudioEngine;
		IXAudio2*							m_pXAudio2;
		Util::CStack<AUDIO_VOICE>			m_sSourceVoiceStack;

		CWaveFile*							m_pWaveFile;
	};

	// Method for checking if this audio source is playing any clips.
	inline bool CXAudio2Source::IsPlaying()
	{
		XAUDIO2_VOICE_STATE state;
		bool played = false;

		// Check for a source voice that is ready to play to sound.
		for(u32 i = 0; i < m_sSourceVoiceStack.Length(); i++)
		{
			m_sSourceVoiceStack[i].voice->GetState(&state);
			if(state.BuffersQueued)
			{
				return true;
			}
		}

		return false;
	}

	// 3D sound getter and setter.

	inline bool CXAudio2Source::Sound3D() const
	{
		return m_b3dSound;
	}

	inline void CXAudio2Source::Sound3D(bool _3dSound)
	{
		m_b3dSound = _3dSound;
	}

	// Distance scale getter and setter.

	inline float CXAudio2Source::DistanceScale() const
	{
		return m_distanceScale;
	}

	inline void CXAudio2Source::DistanceScale(float distanceScale)
	{
		m_distanceScale = distanceScale;
	}

	// Inner cone getter and setters.

	inline float CXAudio2Source::InnerConeAngle() const
	{
		return m_innerCone;
	}

	inline void CXAudio2Source::InnerConeAngle(float innerCone)
	{
		m_innerCone = innerCone;
	}
	
	// Outer cone getter and setters.

	inline float CXAudio2Source::OuterConeAngle() const
	{
		return m_outerCone;
	}

	inline void CXAudio2Source::OuterConeAngle(float outerCone)
	{
		m_outerCone = outerCone;
	}
};

#endif