//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CFBXLoader.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CFBXLoader.h"

FbxManager* gSdkManager = NULL;

#ifdef IOS_REF
	#undef IOS_REF
#endif

#define IOS_REF (*(gSdkManager->GetIOSettings()))

namespace GFX
{
	CFBXLoader::CFBXLoader()
	{
		m_scale								= 0.1f;

		m_pScene							= NULL;

		gSdkManager = FbxManager::Create();
		FbxIOSettings* ios = FbxIOSettings::Create(gSdkManager, IOSROOT);
		gSdkManager->SetIOSettings(ios);
	}

	CFBXLoader::CFBXLoader(const CFBXLoader& other)
	{
	}

	CFBXLoader::~CFBXLoader()
	{
	}

	// Method for loading a fbx file, and storing its contants in a mesh scene object at a specified scale.
	int CFBXLoader::LoadFBXFile(char* filename, CMeshScene*& meshScene, float scale)
	{
		int retCode = Util::Err::NONE;

		m_scale = scale;

		m_pMeshScene = meshScene;
		m_pScene = FbxScene::Create(gSdkManager, "");

		if((retCode = LoadScene(filename)) != Util::Err::NONE)
			return retCode;

		if((retCode = ParseNode(m_pScene->GetRootNode())) != Util::Err::NONE)
			return retCode;

		return retCode;
	}

	// Method for loading the contents of a scene.
	int CFBXLoader::LoadScene(char* filename)
	{
		int retCode = Util::Err::NONE;

		int lFileMajor, lFileMinor, lFileRevision;
		int lSDKMajor,  lSDKMinor,  lSDKRevision;
		int i, lAnimStackCount;

		FbxImporter* pImporter = FbxImporter::Create(gSdkManager, "");

		if(!pImporter->Initialize(filename, -1, gSdkManager->GetIOSettings()))
		{
			return Util::Err::ERR_FBX_LOADER_FAILED_TO_INITIALIZE_FBX_IMPORTER;
		}

		// Get sdk and file versions.
		FbxManager::GetFileFormatVersion(lSDKMajor, lSDKMinor, lSDKRevision);
		pImporter->GetFileVersion(lFileMajor, lFileMinor, lFileRevision);

		if(pImporter->IsFBX())
		{
			lAnimStackCount = pImporter->GetAnimStackCount();

			for(i = 0; i < lAnimStackCount; i++)
			{
				FbxTakeInfo* pTakeInfo = pImporter->GetTakeInfo(i);
			}

			IOS_REF.SetBoolProp(IMP_FBX_MATERIAL,			true);
			IOS_REF.SetBoolProp(IMP_FBX_TEXTURE,			true);
			IOS_REF.SetBoolProp(IMP_FBX_LINK,				true);
			IOS_REF.SetBoolProp(IMP_FBX_SHAPE,				true);
			IOS_REF.SetBoolProp(IMP_FBX_GOBO,				true);
			IOS_REF.SetBoolProp(IMP_FBX_ANIMATION,			true);
			IOS_REF.SetBoolProp(IMP_FBX_GLOBAL_SETTINGS,	true);
		}

		if(!pImporter->Import(m_pScene))
		{
			return Util::Err::ERR_FBX_LOADER_FAILED_TO_IMPORT_SCENE;
		}

		pImporter->Destroy();
		
		return retCode;
	}

	// Method for parsing the data from a fbx file node, and from all of the nodes children.
	int CFBXLoader::ParseNode(FbxNode* node)
	{
		int retCode = Util::Err::NONE;

		if(node)
		{
			if(node->GetMesh())
			{
				if((retCode = ParseMesh(node, node->GetMesh())) != Util::Err::NONE)
					return retCode;
			}

			// Parse child nodes.
			for(int i = 0; i < node->GetChildCount(); i++)
			{
				ParseNode(node->GetChild(i));
			}
		}

		return retCode;
	}

	// Method for parsing a mesh from a fbx file node, and storing all of its data in a a mesh object.
	int CFBXLoader::ParseMesh(FbxNode* node, FbxMesh* mesh)
	{
		int retCode = Util::Err::NONE;
		
		int layerCount, vertexColorCount, uvSetCount;
		int polyCount, vertexCount, polySize, triangleCount;
		int cornerIndices[3];
		int materialIndex, materialCount;
		int i, j, cornerIndex, uvCounter;

		char* name = NULL;

		FbxLayerElementVertexColor*			vertexColorSet		= NULL;
		FbxLayerElementMaterial*			materialSet			= NULL;
		Util::CStack<FbxLayerElementUV*>	vertexUVSets;

		// Store the name of this mesh.
		CMesh* gameMesh = new CMesh();

		gameMesh->m_name = (char*)node->GetName();

		FbxAMatrix meshMtx = CalculateGlobalTransform(node);

		FbxDouble3 geoPosition = node->GeometricTranslation.Get();
		FbxDouble3 geoScale = node->GeometricScaling.Get();
		FbxDouble3 geoRotation = node->GeometricRotation.Get();
		FbxVector4 position = meshMtx.GetT();
		FbxVector4 rotation = meshMtx.GetR();
		FbxVector4 scale = meshMtx.GetS();

		// Store the transformations for this mesh.
		gameMesh->m_pTransform = Math::CTransformSharedPtr(new Math::CTransform());
		gameMesh->m_pTransform->Position(Math::CVector3(-(float)position[0], (float)position[1], (float)position[2]) * m_scale);
		gameMesh->m_pTransform->Euler(Math::CVector3((float)rotation[0], -(float)rotation[1], -(float)rotation[2]) * Math::_PiOver180);
		gameMesh->m_pTransform->Scale(Math::CVector3((float)scale[0], (float)scale[1], (float)scale[2]));
		gameMesh->m_pTransform->Update();

		// Layers.
		layerCount = mesh->GetLayerCount();

		vertexColorCount = 0;
		uvSetCount = 0;

		for(i = 0; i < layerCount; i++)
		{
			// Layer vertex colors.
			if(mesh->GetLayer(i)->GetVertexColors())
			{
				vertexColorCount++;
				vertexColorSet = mesh->GetLayer(i)->GetVertexColors();
			}

			// Layer UVs.
			if(mesh->GetLayer(i)->GetUVs())
			{
				uvSetCount++;
				vertexUVSets.Push(mesh->GetLayer(i)->GetUVs());
			}

			// Layer materials.
			if(mesh->GetLayer(i)->GetMaterials())
			{
				materialSet = mesh->GetLayer(i)->GetMaterials();
				materialCount = materialSet->mDirectArray->GetCount();

				for(j = 0; j < materialCount; j++)
				{
					if((retCode = ParseMaterialInLayer(mesh, gameMesh, mesh->GetLayer(i), j)) != Util::Err::NONE)
						return retCode;
				}
			}
		}

		polyCount = mesh->GetPolygonCount();
		vertexCount = mesh->GetControlPointsCount();
		FbxVector4* vertexPositions = mesh->GetControlPoints();

		// Extract the meshes poly data.
		for(i = 0; i < polyCount; i++)
		{
			polySize = mesh->GetPolygonSize(i);
			triangleCount = polySize - 2;
			if(triangleCount <= 0)
				continue;

			materialIndex = 0;
			if(materialSet)
			{
				switch(materialSet->GetMappingMode())
				{
					case FbxLayerElement::eByPolygon:
						{
							switch(materialSet->GetReferenceMode())
							{
								case FbxLayerElement::eDirect:
									materialIndex = i;
									break;
								case FbxLayerElement::eIndexToDirect:
									materialIndex = materialSet->mIndexArray->GetAt(i);
									break;
								default:
									break;
							}
						}
					default:
						break;
				}
			}

			// Extrant to polygons triangle data.
			for(j = 0; j < triangleCount; j++)
			{
				cornerIndices[0] = mesh->GetPolygonVertex(i, 0);
				cornerIndices[1] = mesh->GetPolygonVertex(i, j + 1);
				cornerIndices[2] = mesh->GetPolygonVertex(i, j + 2);

				FbxVector4 normals[3];
				memset(normals, 0, 3 * sizeof(FbxVector4));
				int vertIndex[] = { 0, j + 1, j + 2 };
				mesh->GetPolygonVertexNormal(i, vertIndex[0], normals[0]);
				mesh->GetPolygonVertexNormal(i, vertIndex[1], normals[1]);
				mesh->GetPolygonVertexNormal(i, vertIndex[2], normals[2]);

				ZeroMemory(&m_triangle, sizeof(FBX_TRIANGLE));

				m_triangle.polygonIndex		= i;
				m_triangle.subsetIndex		= materialIndex;

				for(cornerIndex = 0; cornerIndex < 3; cornerIndex++)
				{
					const int dccIndex							= cornerIndices[cornerIndex];
					m_triangle.dccVertexIndex[cornerIndex]		= dccIndex;
					
					// Position.
					m_triangle.point[cornerIndex].x =  -((float)vertexPositions[dccIndex][0] * (float)geoScale[0] + (float)geoPosition[0]) * m_scale;
					m_triangle.point[cornerIndex].y =  ((float)vertexPositions[dccIndex][1] * (float)geoScale[1] + (float)geoPosition[1]) * m_scale;
					m_triangle.point[cornerIndex].z =  ((float)vertexPositions[dccIndex][2] * (float)geoScale[2] + (float)geoPosition[2]) * m_scale;
					
					Math::CMatrix4x4 geoMtx;
					geoMtx = Math::CMatrix4x4::RotateZ(-(float)geoRotation[2] * Math::_PiOver180);
					geoMtx *= Math::CMatrix4x4::RotateY(-(float)geoRotation[1] * Math::_PiOver180);
					geoMtx *= Math::CMatrix4x4::RotateX((float)geoRotation[0] * Math::_PiOver180);
					
					m_triangle.point[cornerIndex] = geoMtx.TransformCoord(m_triangle.point[cornerIndex]);

					// Normal.
					m_triangle.normal[cornerIndex].x = -(float)normals[cornerIndex][0];
					m_triangle.normal[cornerIndex].y = (float)normals[cornerIndex][1];
					m_triangle.normal[cornerIndex].z = (float)normals[cornerIndex][2];

					for(uvCounter = 0; uvCounter < uvSetCount; uvCounter++)
					{
						FbxLayerElementUV* uvSet = (FbxLayerElementUV*)vertexUVSets[uvCounter];
						FbxVector2 value(0, 0);
						int uvIndex = 0;

						switch(uvSet->GetMappingMode())
						{
							case FbxLayerElement::eByControlPoint:
								uvIndex = mesh->GetPolygonVertex(i, vertIndex[cornerIndex]);
								break;
							case FbxLayerElement::eByPolygonVertex:
								uvIndex = mesh->GetTextureUVIndex(i, vertIndex[cornerIndex]);
								break;
							default:
								break;
						}

						value = uvSet->mDirectArray->GetAt(uvIndex);
						m_triangle.uv[cornerIndex][uvCounter].x = (float)value[0];
						m_triangle.uv[cornerIndex][uvCounter].y = -(float)value[1];
					}
					
					if(vertexColorCount > 0 && vertexColorSet)
					{
						FbxColor value(0, 0, 0, 1);

						switch(vertexColorSet->GetMappingMode())
						{
							case FbxLayerElement::eByControlPoint:
								switch(vertexColorSet->GetReferenceMode())
								{
									case FbxLayerElement::eDirect:
										value = vertexColorSet->mDirectArray->GetAt(dccIndex);
										break;
									case FbxLayerElement::eIndexToDirect:
										value = vertexColorSet->mDirectArray->GetAt(vertexColorSet->mIndexArray->GetAt(dccIndex));
										break;
									default:
										break;
								}
								break;
							case FbxLayerElement::eByPolygonVertex:
								switch(vertexColorSet->GetReferenceMode())
								{
									case FbxLayerElement::eDirect:
										value = vertexColorSet->mDirectArray->GetAt(vertIndex[cornerIndex]);
										break;
									case FbxLayerElement::eIndexToDirect:
										value = vertexColorSet->mDirectArray->GetAt(vertexColorSet->mIndexArray->GetAt(vertIndex[cornerIndex]));
										break;
									default:
										break;
								}
								break;
							default:
								break;
						}

						m_triangle.color[cornerIndex].r = (float)value.mRed;
						m_triangle.color[cornerIndex].g = (float)value.mGreen;
						m_triangle.color[cornerIndex].b = (float)value.mBlue;
						m_triangle.color[cornerIndex].a = (float)value.mAlpha;
					}
				}
				
				m_triangle.CalculateTangentBinormal();

				gameMesh->AddTriangle(
					new CMesh::MESH_VERTEX(m_triangle.point[0], m_triangle.normal[0], m_triangle.binormal[0], m_triangle.tangent[0], m_triangle.color[0], m_triangle.uv[0]),
					new CMesh::MESH_VERTEX(m_triangle.point[1], m_triangle.normal[1], m_triangle.binormal[1], m_triangle.tangent[1], m_triangle.color[1], m_triangle.uv[1]),
					new CMesh::MESH_VERTEX(m_triangle.point[2], m_triangle.normal[2], m_triangle.binormal[2], m_triangle.tangent[2], m_triangle.color[2], m_triangle.uv[2]),
					true
				);
			}
		}

		m_pMeshScene->PushMesh(gameMesh);

		return retCode;
	}
	
	// Method for parsing a material from a fbx mesh.
	int CFBXLoader::ParseMaterialInLayer(FbxMesh* mesh, CMesh* gameMesh, FbxLayer* layer, int materialIndex)
	{
		int retCode = Util::Err::NONE;

		if(materialIndex >= layer->GetMaterials()->mDirectArray->GetCount())
			return Util::Err::ERR_FBX_LOADER_INVALID_MATERIAL_INDEX;

		FbxSurfaceMaterial* fbxMaterial = mesh->GetNode()->GetMaterial(materialIndex);

		// Check if the material is already stored in the game mesh scene material list.
		CMatMesh* material = FindMaterial(fbxMaterial->GetUniqueID());
		if(material)
		{
			gameMesh->m_sMaterialStack.Push(material);
			return retCode;
		}

		// Check if the material should be renamed.
		std::string materialName = fbxMaterial->GetName();
		int index = 0;
		do
		{
			material = FindMaterial(materialName);

			if(material)
			{
				materialName = materialName + "_" + std::to_string(index++);
			}

		} while(material);

		material							= new CMatMesh();
		material->m_uniqueId				= fbxMaterial->GetUniqueID();
		material->m_name					= materialName;
		material->m_shadingModel			= SHADING_MODEL::UNKNOWN;

		// Shading models.
		if(fbxMaterial->Is<FbxSurfacePhong>())
		{
			material->m_shadingModel = SHADING_MODEL::PHONG;

			FbxDouble3 color							= ((FbxSurfacePhong*)fbxMaterial)->Ambient;
			material->m_colorBuffer.ambient				= Math::CColor((float)color[0], (float)color[1], (float)color[2]);

			color										= ((FbxSurfacePhong*)fbxMaterial)->Diffuse;
			material->m_colorBuffer.diffuse				= Math::CColor((float)color[0], (float)color[1], (float)color[2]);

			color										= ((FbxSurfacePhong*)fbxMaterial)->Specular;
			material->m_colorBuffer.specular			= Math::CColor((float)color[0], (float)color[1], (float)color[2]);

			color										= ((FbxSurfacePhong*)fbxMaterial)->Emissive;
			material->m_colorBuffer.emissive			= Math::CColor((float)color[0], (float)color[1], (float)color[2]);

			material->m_colorBuffer.transparency		= 1.0f - (float)((FbxSurfacePhong*)fbxMaterial)->TransparencyFactor.Get();
			material->m_lightingBuffer.specular			= (float)((FbxSurfacePhong*)fbxMaterial)->SpecularFactor.Get();
			material->m_lightingBuffer.shininess		= (float)((FbxSurfacePhong*)fbxMaterial)->Shininess.Get();
			material->m_lightingBuffer.reflectivity		= (float)((FbxSurfacePhong*)fbxMaterial)->ReflectionFactor.Get();
		}
		else if(fbxMaterial->Is<FbxSurfacePhong>())
		{
			material->m_shadingModel = SHADING_MODEL::LAMBERT;

			FbxDouble3 color							= ((FbxSurfaceLambert*)fbxMaterial)->Ambient;
			material->m_colorBuffer.ambient				= Math::CColor((float)color.mData[0], (float)color.mData[1], (float)color.mData[2]);

			color										= ((FbxSurfaceLambert*)fbxMaterial)->Diffuse;
			material->m_colorBuffer.diffuse				= Math::CColor((float)color.mData[0], (float)color.mData[1], (float)color.mData[2]);

			color										= ((FbxSurfaceLambert*)fbxMaterial)->Emissive;
			material->m_colorBuffer.emissive			= Math::CColor((float)color.mData[0], (float)color.mData[1], (float)color.mData[2]);

			material->m_colorBuffer.transparency		= 1.0f - (float)((FbxSurfaceLambert*)fbxMaterial)->TransparencyFactor.Get();
		}

		// Perp for texture extraction.
		enum POST_OP
		{
			PO_NOTHING,
			PO_TRANSPARENT,
		};

		struct TEXTURE_PARAM_EXTRACTION
		{
			const char*	propertyName;
			const char*	parameterName;
			POST_OP		postOp;
			u32			parameterFlag;
		};

		TEXTURE_PARAM_EXTRACTION ExtractionList[] =
		{
			{ FbxSurfaceMaterial::sTransparentColor,		"AlphaMapTexture",				PO_TRANSPARENT,			MESH_MAT_PARAM::MMP_ALPHA_MAP		},
			{ FbxSurfaceMaterial::sDiffuse,					"DiffuseMapTexture",			PO_NOTHING,				MESH_MAT_PARAM::MMP_DIFFUSE_MAP		},
			{ FbxSurfaceMaterial::sBump,					"NormalMapTexture",				PO_NOTHING,				MESH_MAT_PARAM::MMP_BUMP_MAP		},
			{ FbxSurfaceMaterial::sSpecular,				"SpecularMapTexture",			PO_NOTHING,				MESH_MAT_PARAM::MMP_SPECULAR_MAP	},
			{ FbxSurfaceMaterial::sEmissive,				"EmissiveMapTexture",			PO_NOTHING,				MESH_MAT_PARAM::MMP_EMISSIVE_MAP	},
		};

		// Extract textures from the material.
		for(int i = 0; i < ARRAYSIZE(ExtractionList); i++)
		{
			TEXTURE_PARAM_EXTRACTION tpe = ExtractionList[i];
			FbxProperty prop = fbxMaterial->FindProperty(tpe.propertyName);
			
			if(!prop.IsValid())
				continue;

			ExtractTextures(prop, tpe.parameterName, material, tpe.parameterFlag);
			
			if(tpe.postOp & PO_TRANSPARENT)
			{
				material->m_bTransparent = true;
			}
		}

		// Add the material to the mesh scene material list.
		m_pMeshScene->m_sMaterialList.Push(material);
		gameMesh->m_sMaterialStack.Push(material);

		return retCode;
	}

	// Method for extracting the textures from a material.
	int CFBXLoader::ExtractTextures(FbxProperty prop, const char* paramName, CMatMesh* material, u32 flags)
	{
		int retCode = Util::Err::NONE;

		int layeredTextureCount = prop.GetSrcObjectCount<FbxLayeredTexture>();
		
		int i;
		int index = 0;
		int textureCount = 0;

		if(layeredTextureCount > 0)
		{
			int j;

			FbxFileTexture* fbxTexture;
			FbxLayeredTexture* fbxLayeredTexture;

			// Extract textures from a layered texture array.
			for(i = 0; i < layeredTextureCount; i++)
			{
				fbxLayeredTexture = prop.GetSrcObject<FbxLayeredTexture>(i);
				textureCount = fbxLayeredTexture->GetSrcObjectCount<FbxLayeredTexture>();

				for(j = 0; j < textureCount; j++)
				{
					fbxTexture = FbxCast<FbxFileTexture>(fbxLayeredTexture->GetSrcObject<FbxLayeredTexture>(j));
					if(fbxTexture == NULL)
						continue;

					AddTextureParameter(material, paramName, index, fbxTexture->GetRelativeFileName(), flags);
					index++;
				}
			}
		}
		else
		{
			FbxFileTexture* fbxTexture;

			textureCount = prop.GetSrcObjectCount<FbxFileTexture>();

			// Extract the textures from a texture file array.
			for(i = 0; i < textureCount; i++)
			{
				fbxTexture = FbxCast<FbxFileTexture>(prop.GetSrcObject<FbxFileTexture>(i));
				if(fbxTexture == NULL)
					continue;

				AddTextureParameter(material, paramName, index, fbxTexture->GetRelativeFileName(), flags);
				index++;
			}
		}

		material->m_textureCount = index;

		return retCode;
	}

	// Method for add the texture parameter data to a texture.
	void CFBXLoader::AddTextureParameter(CMatMesh* material, const char* paramName, int index, const char* filename, u32 flags)
	{
		CMeshMaterialParameter outputParam;

		if(index == 0)
		{
			strcpy_s(outputParam.name, paramName);
		}
		else
		{
			strcpy_s(outputParam.name, (std::string(paramName) + std::to_string(index)).c_str());
		}

		strcpy_s(outputParam.valString, filename);
		outputParam.paramType			= MPT_TEXTURE2D;
		outputParam.bInstanceParam		= true;
		outputParam.flags				= flags;
		material->m_sMaterialParamStack.Push(outputParam);
	}

	// Method for calculating an fbx file node's world transformation matrix.
	FbxAMatrix CFBXLoader::CalculateGlobalTransform(FbxNode* node)
	{
		FbxAMatrix lTranlationM, lScalingM, lScalingPivotM, lScalingOffsetM, lRotationOffsetM, lRotationPivotM,
					lPreRotationM, lRotationM, lPostRotationM, lTransform;

		FbxAMatrix lParentGX, lGlobalT, lGlobalRS;

		if(!node)
		{
			lTransform.SetIdentity();
			return lTransform;
		}

		// Translation.
		FbxVector4 lTranslation = node->LclTranslation.Get();
		lTranlationM.SetT(lTranslation);

		// Rotation.
		FbxVector4 lRotation = node->LclRotation.Get();
		FbxVector4 lPreRotation = node->PreRotation.Get();
		FbxVector4 lPostRotation = node->PostRotation.Get();

		lRotationM.SetR(lRotation);
		lPreRotationM.SetR(lPreRotation);
		lPostRotationM.SetR(lPostRotation);

		// Scaling.
		FbxVector4 lScale = node->LclScaling.Get();
		lScalingM.SetS(lScale);

		// Pivots and offsets.
		FbxVector4 lScalingOffset = node->ScalingOffset.Get();
		FbxVector4 lScalingPivot = node->ScalingPivot.Get();
		FbxVector4 lRotationOffset = node->RotationOffset.Get();
		FbxVector4 lRotationPivot = node->RotationPivot.Get();
		lScalingOffsetM.SetT(lScalingOffset);
		lScalingPivotM.SetT(lScalingPivot);
		lRotationOffsetM.SetT(lRotationOffset);
		lRotationPivotM.SetT(lRotationPivot);

		// Calculate parent world.
		FbxNode* lParentNode = node->GetParent();
		if(lParentNode)
		{
			lParentGX = CalculateGlobalTransform(lParentNode);
		}
		else
		{
			lParentGX.SetIdentity();
		}

		// Global rotaion.
		FbxAMatrix lLRM, lParentGRM;
		FbxVector4 lParentGR = lParentGX.GetR();
		lParentGRM.SetR(lParentGR);
		lLRM = lPreRotationM * lRotationM * lPostRotationM;

		// Global Scale.
		FbxAMatrix lLSM, lParentGSM, lParentGRSM, lParentTM;
		FbxVector4 lParentGT = lParentGX.GetT();
		lParentTM.SetT(lParentGT);
		lParentGRSM = lParentTM.Inverse() * lParentGX;
		lParentGSM = lParentGRM.Inverse() * lParentGRSM;
		lLSM = lScalingM;

		// Handle inheritance type.
		FbxTransform::EInheritType lInheritType = node->InheritType.Get();
		if(lInheritType == FbxTransform::eInheritRrSs)
		{
			lGlobalRS = lParentGRM * lLRM * lParentGSM * lLSM;
		}
		else if(lInheritType == FbxTransform::eInheritRSrs)
		{
			lGlobalRS = lParentGRM * lParentGSM * lLRM * lLSM;
		}
		else if(lInheritType == FbxTransform::eInheritRrs)
		{
			FbxAMatrix lParentLSM;
			FbxVector4 lParentLS = lParentNode->LclScaling.Get();
			lParentLSM.SetS(lParentLS);

			FbxAMatrix lParentGSM_noLocal = lParentGSM * lParentLSM.Inverse();
			lGlobalRS = lParentGRM * lLRM * lParentGSM_noLocal * lLSM;
		}
		else
		{
			// Unknown inheritance type. Use RrSs as a substitute.
			lGlobalRS = lParentGRM * lLRM * lParentGSM * lLSM;
		}

		// Global Translation.
		lTransform = lTranlationM * lRotationOffsetM * lRotationPivotM * lPreRotationM * lRotationM * lPostRotationM * lRotationPivotM.Inverse() *
			lScalingOffsetM * lScalingPivotM * lScalingM * lScalingPivotM.Inverse();

		FbxVector4 lLocalTWithAllPivotAndOffsetInfo = lTransform.GetT();

		FbxVector4 lGlobalTranslation = lParentGX.MultT(lLocalTWithAllPivotAndOffsetInfo);
		lGlobalT.SetT(lGlobalTranslation);

		//Construct the whole global transform.
		lTransform = lGlobalT * lGlobalRS;

		return lTransform;
	}
	
	// Method for finding a loaded material a id.
	CMatMesh* CFBXLoader::FindMaterial(u64 id)
	{
		for(u32 i = 0; i < m_pMeshScene->m_sMaterialList.Length(); i++)
		{
			if(m_pMeshScene->m_sMaterialList[i] && ((CMatMesh*)m_pMeshScene->m_sMaterialList[i])->m_uniqueId == id)
			{
				return (CMatMesh*)m_pMeshScene->m_sMaterialList[i];
			}
		}

		return NULL;
	}

	// Method for finding a loaded material a name.
	CMatMesh* CFBXLoader::FindMaterial(const std::string& name)
	{
		for(u32 i = 0; i < m_pMeshScene->m_sMaterialList.Length(); i++)
		{
			if(m_pMeshScene->m_sMaterialList[i] && ((CMatMesh*)m_pMeshScene->m_sMaterialList[i])->m_name == name)
			{
				return (CMatMesh*)m_pMeshScene->m_sMaterialList[i];
			}
		}

		return NULL;
	}

	// Method for releasing the fbx data.
	void CFBXLoader::Release()
	{
		if(m_pScene)
			m_pScene->Destroy();
		if(gSdkManager)
			gSdkManager->Destroy();
	}
};