//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CLighting.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CLIGHTING_H
#define CLIGHTING_H

#include "CConstantBuffer.h"
#include "..\Math\CMathVector2.h"
#include "..\Math\CMathCamera.h"
#include "CTextureRenderTo.h"
#include "CRenderingData.h"
#include "CMatShadowMap.h"
#include "CDebugSphere.h"
#include "..\Utilities\CDataQueue.h"

#include "CLightDirectional.h"
#include "CLightPoint.h"
#include "CLightSpot.h"

namespace GFX
{
	// CLighting class.
	// Class for creating and rendering (through deferred rendering), directional lights, point lights, and spotlights.
	// Handles dynamic shadow mapping for each type of supported light as well.
	class CLighting : public App::CDX11Object
	{
	public:
		friend class CMatPostScreen;
		
	public:
		CLighting();
		CLighting(const CLighting& other);
		virtual ~CLighting();

		int									Initialize();
		int									Finalize();

		int									ResetBuffer();
		int									BindBuffer();

		int									RenderMaps(void* data, int (*RenderWorldObjects)(void*, CMaterial*));
		int									Render();
		
		void								Release();

		inline bool							DebugMode() const;
		inline void							DebugMode(bool debugMode);

		inline void							AddDirectionalLight(DIRECTIONAL_LIGHT_QUEUE* dirLight);
		inline void							AddPointLight(POINT_LIGHT_QUEUE* pointLight);
		inline void							AddSpotlight(SPOTLIGHT_QUEUE* spotlight);

		inline RENDERDATA*					RenderData() const;
		inline void							RenderData(RENDERDATA* pRenderData);

		inline CMatDebug*					DebugMaterial() const;
		inline void							DebugMaterial(CMatDebug* pMatDebug);
		inline Math::CCamera*				CurrentCamera() const;
		inline void							CurrentCamera(Math::CCamera* pCurrentCamera);

	private:
		CDebugSphere*						m_pDebugSphere;

		bool								m_bDebugMode;
		u32									m_shadowMapSize;
		
		CMatShadowMap*						m_pMatShadowMap;
		RENDERDATA*							m_pRenderData;

		CLightDirectional*					m_pLightDirectional;
		CLightPoint*						m_pLightPoint;
		CLightSpot*							m_pLightSpot;

		CMatDebug*							m_pMatDebug;
		Math::CCamera*						m_pCurrentCamera;
	};

	// Debug mode getter and setter.

	inline bool CLighting::DebugMode() const
	{
		return m_bDebugMode;
	}

	inline void CLighting::DebugMode(bool debugMode)
	{
		m_bDebugMode = debugMode;
	}

	// Method for adding a direction light to the lighting system.
	inline void CLighting::AddDirectionalLight(DIRECTIONAL_LIGHT_QUEUE* dirLight)
	{
		m_pLightDirectional->AddDirectionalLight(dirLight);
	}
	
	// Method for adding a point light to the lighting system.
	inline void CLighting::AddPointLight(POINT_LIGHT_QUEUE* pointLight)
	{
		m_pLightPoint->AddPointLight(pointLight);
	}
	
	// Method for adding a spotlight to the lighting system.
	inline void CLighting::AddSpotlight(SPOTLIGHT_QUEUE* spotLight)
	{
		m_pLightSpot->AddSpotlight(spotLight);
	}

	// Render data getter and setter.

	inline RENDERDATA* CLighting::RenderData() const
	{
		return m_pRenderData;
	}

	inline void CLighting::RenderData(RENDERDATA* pRenderData)
	{
		m_pRenderData = pRenderData;
	}

	// Debug material getter and setter.

	inline CMatDebug* CLighting::DebugMaterial() const
	{
		return m_pMatDebug;
	}

	inline void CLighting::DebugMaterial(CMatDebug* pMatDebug)
	{
		m_pMatDebug = pMatDebug;
	}

	// Current camera getter and setter.

	inline Math::CCamera* CLighting::CurrentCamera() const
	{
		return m_pCurrentCamera;
	}

	inline void CLighting::CurrentCamera(Math::CCamera* pCurrentCamera)
	{
		m_pCurrentCamera = pCurrentCamera;
	}
};

#endif
