//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CMaterial.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CMATERIAL_H
#define CMATERIAL_H

#include <memory>
#include "CShader.h"
#include "..\Math\CMathTransform.h"
#include "..\Math\CMathCamera.h"
#include "CTexture.h"
#include "CRenderingData.h"

namespace GFX
{
	// Regular matrix constant buffer.
	struct MatrixCBuffer
	{
		Math::CMatrix4x4					world;
		Math::CMatrix4x4					view;
		Math::CMatrix4x4					projection;
	};

	// Matrix constant buffer w/ inverse matices.
	struct MatrixInvCBuffer
	{
		Math::CMatrix4x4					world;
		Math::CMatrix4x4					worldInv;
		Math::CMatrix4x4					view;
		Math::CMatrix4x4					viewInv;
		Math::CMatrix4x4					projection;
		Math::CMatrix4x4					projectionInv;
	};

	class CMaterial;

	typedef std::shared_ptr<CMaterial> CMaterialSharedPtr;
	typedef std::unique_ptr<CMaterial> CMaterialPtr;

	// CMaterial class.
	// Base class for creating new materials that can be used with custom shaders, and geometry.
	class CMaterial : public App::CDX11Object
	{
	public:
		CMaterial();
		CMaterial(const CMaterial& other);
		virtual ~CMaterial();

		virtual int CMaterial::Initialize();
		virtual int CMaterial::Render();
		virtual void CMaterial::Release();

		inline u32							TextureCount() const;
		inline void							TextureCount(u32 textureCount);

		inline RENDERDATA*					RenderData() const;
		inline void							RenderData(RENDERDATA* pRenderData);

		inline CTexture*					Texture() const;
		inline void							Texture(CTexture* pTexture);

		inline CTexture**					TextureArray() const;
		inline void							TextureArray(CTexture** ppTexture);
		inline void							TextureArray(u32 index, CTexture* pTexture);

		inline Math::CCamera*				Camera();
		inline void							Camera(Math::CCamera* pCamera);
		inline Math::CTransformSharedPtr	Transform() const;
		inline void							Transform(Math::CTransformSharedPtr pTransform);

		inline D3D_SHADER_MACRO*			ShaderMacros() const;
		inline void							ShaderMacros(D3D_SHADER_MACRO* pMacro);

	protected:
		virtual int							SetupShader() { return Util::Err::NONE; }

		D3D11_INPUT_ELEMENT_DESC*			m_pIed;
		u32									m_elementCount;
		u32									m_bufferCount;
		u32									m_textureCount;

		char*								m_pVertexResName;
		char*								m_pHullResName;
		char*								m_pDomainResName;
		char*								m_pPixelResName;
		char*								m_pGeometryResName;
		char*								m_pComputeResName;

		char*								m_pVertexEntryPoint;
		char*								m_pHullEntryPoint;
		char*								m_pDomainEntryPoint;
		char*								m_pPixelEntryPoint;
		char*								m_pGeometryEntryPoint;
		char*								m_pComputeEntryPoint;

		bool								m_bUseInverse;

		SHADER_TYPE							m_shaderType;
		MatrixCBuffer						m_mtxCBuffer;
		MatrixInvCBuffer					m_mtxInvCBuffer;

		char*								m_pLogFilename;

		D3D_SHADER_MACRO*					m_pMacros;

		CTexture*							m_pTexture;
		CTexture**							m_pTextureArray;

		RENDERDATA*							m_pRenderData;
		CShader*							m_pShader;
		Math::CCamera*						m_pCamera;
		Math::CTransformSharedPtr			m_pTransform;
	};

	// Texture count getter and setter.

	inline u32 CMaterial::TextureCount() const
	{
		return m_textureCount;
	}

	inline void CMaterial::TextureCount(u32 textureCount)
	{
		m_textureCount = textureCount;
	}

	// Render data getter and setter.

	inline RENDERDATA* CMaterial::RenderData() const
	{
		return m_pRenderData;
	}

	inline void CMaterial::RenderData(RENDERDATA* pRenderData)
	{
		m_pRenderData = pRenderData;
	}

	// Texture getter and setter.

	inline CTexture* CMaterial::Texture() const
	{
		return m_pTexture;
	}

	inline void CMaterial::Texture(CTexture* pTexture)
	{
		m_pTexture = pTexture;
	}

	// Texture array getters and setters.

	inline CTexture** CMaterial::TextureArray() const
	{
		return m_pTextureArray;
	}

	inline void CMaterial::TextureArray(CTexture** ppTexture)
	{
		m_pTextureArray = ppTexture;
	}

	inline void CMaterial::TextureArray(u32 index, CTexture* pTexture)
	{
		m_pTextureArray[index] = pTexture;
	}
	
	// Camera getter and setter.

	inline Math::CCamera* CMaterial::Camera()
	{
		return m_pCamera;
	}

	inline void CMaterial::Camera(Math::CCamera* pCamera)
	{
		m_pCamera = pCamera;
	}
	
	// Transform getter and setter.

	inline Math::CTransformSharedPtr CMaterial::Transform() const
	{
		return m_pTransform;
	}

	inline void CMaterial::Transform(Math::CTransformSharedPtr pTransform)
	{
		m_pTransform = pTransform;
	}
	
	// Shader macros getter and setter.

	inline D3D_SHADER_MACRO* CMaterial::ShaderMacros() const
	{
		return m_pMacros;
	}

	inline void CMaterial::ShaderMacros(D3D_SHADER_MACRO* pMacro)
	{
		m_pMacros = pMacro;
	}
};

#endif