//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CMeshScene.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CMeshScene.h"

namespace GFX
{
	CMeshScene::CMeshScene()
	{
		m_pDebugMaterial					= NULL;

		m_pRenderData						= NULL;
		m_pCamera							= NULL;
	}

	CMeshScene::CMeshScene(const CMeshScene& other)
	{
	}

	CMeshScene::~CMeshScene()
	{
	}

	// Method for initializing the mesh scene's mesh and material data.
	int CMeshScene::Initialize()
	{
		int retCode = Util::Err::NONE;

		u32 i;

		CMatMesh* meshMat;
		for(i = 0; i < m_sMaterialList.Length(); i++)
		{
			meshMat = ((CMatMesh*)m_sMaterialList[i]);

			meshMat->DirectXData(m_pDX11Data);
			meshMat->RenderData(m_pRenderData);
			meshMat->Camera(m_pCamera);

			if((retCode = meshMat->Initialize()) != Util::Err::NONE)
				return retCode;
		}

		CMesh* mesh;
		
		// Initialize opaque meshes.
		for(i = 0; i < m_sMeshList.Length(); i++)
		{
			mesh = ((CMesh*)m_sMeshList[i]);

			mesh->DirectXData(m_pDX11Data);
			mesh->RenderData(m_pRenderData);
			mesh->DebugMaterial(m_pDebugMaterial);

			if((retCode = mesh->Initialize()) != Util::Err::NONE)
				return retCode;
		}

		// Initialize alpha meshes.
		for(i = 0; i < m_sMeshAlphaList.Length(); i++)
		{
			mesh = ((CMesh*)m_sMeshAlphaList[i]);

			mesh->DirectXData(m_pDX11Data);
			mesh->RenderData(m_pRenderData);
			mesh->DebugMaterial(m_pDebugMaterial);

			if((retCode = mesh->Initialize()) != Util::Err::NONE)
				return retCode;
		}

		return retCode;
	}

	// Method for updating and sorting the scene meshes.
	int CMeshScene::Update()
	{
		int retCode = Util::Err::NONE;
		
		u32 i;
		
		// Update opaque meshes.
		for(i = 0; i < m_sMeshList.Length(); i++)
		{
			if((retCode = ((CMesh*)m_sMeshList[i])->Update()) != Util::Err::NONE)
				return retCode;
		}
		
		// Update alpha meshes.
		for(i = 0; i < m_sMeshAlphaList.Length(); i++)
		{
			if((retCode = ((CMesh*)m_sMeshAlphaList[i])->Update()) != Util::Err::NONE)
				return retCode;
		}

		// Sort alpha meshes.
		auto SortAlphaMeshes = [] (CMesh* a, CMesh* b)
		{
			if(!a->Transform() || !a->MeshMaterial()->Camera() || !a->MeshMaterial()->Camera()->Transform() || 
				!b->Transform() || !b->MeshMaterial()->Camera() || !b->MeshMaterial()->Camera()->Transform())
					return true;

			return (a->Transform()->DistanceFrom(a->MeshMaterial()->Camera()->Transform().get()) < 
				(b->Transform()->DistanceFrom(b->MeshMaterial()->Camera()->Transform().get())));
		};

		Util::QuickSort<CMesh*>(m_sMeshAlphaList.Array(), SortAlphaMeshes, 0, (int)m_sMeshAlphaList.Length() - 1);

		return retCode;
	}
	
	// Method for rendering opaque meshes.
	int CMeshScene::Render()
	{
		int retCode = Util::Err::NONE;

		u32 i;
		
		// Render opaque meshes.
		for(i = 0; i < m_sMeshList.Length(); i++)
		{
			if((retCode = ((CMesh*)m_sMeshList[i])->Render()) != Util::Err::NONE)
				return retCode;
		}

		return retCode;
	}

	// Method for rendering opaque meshes with a custom material.
	int CMeshScene::Render(CMaterial* material)
	{
		int retCode = Util::Err::NONE;

		u32 i;
		
		// Render opaque meshes.
		for(i = 0; i < m_sMeshList.Length(); i++)
		{
			if(material)
			{
				if((retCode = ((CMesh*)m_sMeshList[i])->Render(material)) != Util::Err::NONE)
					return retCode;
			}
			else
			{
				if((retCode = ((CMesh*)m_sMeshList[i])->Render()) != Util::Err::NONE)
					return retCode;
			}
		}

		return retCode;
	}

	// Method for render transparent meshes.
	int CMeshScene::RenderAlpha()
	{
		int retCode = Util::Err::NONE;

		u32 i;
		
		// Render transparent meshes.
		for(i = 0; i < m_sMeshAlphaList.Length(); i++)
		{
			if((retCode = ((CMesh*)m_sMeshAlphaList[i])->Render()) != Util::Err::NONE)
				return retCode;
		}

		return retCode;
	}
	
	// Method for rendering transparent meshes with a custom material.
	int CMeshScene::RenderAlpha(CMaterial* material)
	{
		int retCode = Util::Err::NONE;

		u32 i;
		
		// Render transparent meshes.
		for(i = 0; i < m_sMeshAlphaList.Length(); i++)
		{
			if(material)
			{
				if((retCode = ((CMesh*)m_sMeshAlphaList[i])->Render(material)) != Util::Err::NONE)
					return retCode;
			}
			else
			{
				if((retCode = ((CMesh*)m_sMeshAlphaList[i])->Render()) != Util::Err::NONE)
					return retCode;
			}
		}

		return retCode;
	}

	// Method for releasing all meshes and materials in the scene.
	void CMeshScene::Release()
	{
		u32 i;

		CMesh* mesh;
		
		// Free all alpha meshes in this scene.
		for(i = 0; i < m_sMeshAlphaList.Length(); i++)
		{
			mesh = ((CMesh*)m_sMeshAlphaList[i]);
			SAFE_RELEASE_DELETE(mesh);
		}

		m_sMeshAlphaList.Release();

		// Free all meshes in this scene.
		for(i = 0; i < m_sMeshList.Length(); i++)
		{
			mesh = ((CMesh*)m_sMeshList[i]);
			SAFE_RELEASE_DELETE(mesh);
		}

		m_sMeshList.Release();

		// Free all materials in this scene.
		CMatMesh* meshMat;
		for(i = 0; i < m_sMaterialList.Length(); i++)
		{
			meshMat = ((CMatMesh*)m_sMaterialList[i]);
			SAFE_RELEASE_DELETE(meshMat);
		}

		m_sMaterialList.Release();
	}
	
	// Method for finding a mesh with a name.
	CMesh* CMeshScene::FindMesh(const std::string& name)
	{
		u32 i;

		for(i = 0; i < m_sMeshList.Length(); i++)
		{
			if(m_sMeshList[i]->m_name == name)
			{
				return m_sMeshList[i];
			}
		}

		for(i = 0; i < m_sMeshAlphaList.Length(); i++)
		{
			if(m_sMeshAlphaList[i]->m_name == name)
			{
				return m_sMeshAlphaList[i];
			}
		}

		return NULL;
	}
};