//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Graphics\CParticleDevice.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CPARTICLEDEVICE_H
#define CPARTICLEDEVICE_H

#include "..\Utilities\CDebugError.h"
#include "..\Utilities\CTimer.h"
#include "CGeoObject.h"
#include "CMatParticle.h"

namespace GFX
{
	typedef Math::CColor		ParticleFade		(void* data, float lifetime);
	typedef Math::CVector3		ParticleScale		(void* data, float lifetime);

	// Enumerator for supported particle device playback states.
	enum class PARTICLE_PLAYBACK_STATE
	{
		Play,
		Resume,
		Pause,
		Stop,
		Reset,
	};

	// Enumerator for supported particle billboard types.
	// Billboarding takes place in the particle geometry shader.
	enum PARTICLE_BILLBOARD_TYPE
	{
		PBT_None		= 0x0,
		PBT_X			= 0x1,
		PBT_Y			= 0x2,
		PBT_Z			= 0x4,
		PBT_Flat		= 0x5,
		PBT_All			= 0x7,
	};

	// CParticleDevice class.
	// Class for creating a general particle device class.
	class CParticleDevice : public CGeoObject
	{
	public:
		friend class CParticleSystem;

	public:
		// Particle vertex structure.
		struct PARTICLE_VERTEX
		{
			Math::CVector3					position;

			PARTICLE_VERTEX()
			{
			}

			PARTICLE_VERTEX(Math::CVector3 Position)
			{
				position = Position;
			}
		};

		// Particle instance structure.
		struct PARTICLE_INSTANCE
		{
			Math::CVector3					position;
			Math::CVector3					scale;
			Math::CVector4					uvOffset;
			Math::CColor					color;
			Math::CVector4					data0; // x = startTime, y = lifetime, z = '', w = ''.
		};

	public:
		CParticleDevice();
		CParticleDevice(const CParticleDevice& other);
		virtual ~CParticleDevice();

		inline bool							Loop() const;
		inline void							Loop(bool loop);

		inline float						Duration() const;
		inline void							Duration(float duration);

		inline PARTICLE_PLAYBACK_STATE		PlaybackState() const;
		inline void							PlaybackState(const PARTICLE_PLAYBACK_STATE& playbackState);

		inline float						Lifetime() const;
		inline void							Lifetime(float lifetime);

		inline float						SpawnRate() const;
		inline void							SpawnRate(float spawnRate);
		
		inline Math::CVector3				Velocity() const;
		inline void							Velocity(const Math::CVector3& velocity);

		inline Math::CVector3				InitialScale() const;
		inline void							InitialScale(const Math::CVector3& scale);

		inline Math::CColor					InitialColor() const;
		inline void							InitialColor(const Math::CColor& color);

		inline float						Softness() const;
		inline void							Softness(float softness);
		
		inline PARTICLE_BILLBOARD_TYPE		BillboardType() const;
		inline void							BillboardType(int billboardType);
		
		inline float						PlaybackTime() const;

		inline void							FunctionSetup(void* data);
		inline void							FadeFunction(ParticleFade* particleFade);
		inline void							ScaleFunction(ParticleScale* particleScale);

		inline u32							MaxParticles() const;
		inline void							MaxParticles(u32 maxParictles);

		inline PARTICLE_INSTANCE&			InstanceAt(u32 index);

		inline std::string					TextureFilename() const;
		inline void							TextureFilename(const std::string& filename);

		inline CTexture**					TextureArray() const;
		inline void							TextureArray(CTexture** ppTexture);
		inline CTexture*					Texture(u32 index = 0) const;
		inline void							Texture(CTexture* pTexture, u32 index = 0);

		inline Util::CTimer*				Timer() const;
		inline void							Timer(Util::CTimer* pTimer);

	protected:
		int									Initialize() override;
		int									Update() override;
		int									Render() override;

		virtual int							Simulate();
		virtual void						Spawning();
		virtual void						Reset();

		void								Release() override;

		bool								m_bLoop;

		float								m_duration;
		float								m_playbackTimer;

		float								m_lifetime;
		float								m_spawnTime;
		float								m_spawnTimer;

		float								m_spawnRate;

		Math::CVector3						m_velocity;
		Math::CVector3						m_initialScale;
		
		Math::CColor						m_color;
		float								m_softness;

		u32									m_textureCount;

		void*								m_pParticleFunctionData;
		ParticleFade*						m_ParticleFade;
		ParticleScale*						m_ParticleScale;

		PARTICLE_PLAYBACK_STATE				m_playbackState;
		PARTICLE_BILLBOARD_TYPE				m_billboardType;
		std::string							m_textureFilename;

		Util::CTimer*						m_pTimer;
		CTexture**							m_ppTexture;

	private:
		bool								m_bTextureMember;
	};

	// Particle looping getter and setter..

	inline bool CParticleDevice::Loop() const
	{
		return m_bLoop;
	}

	inline void CParticleDevice::Loop(bool loop)
	{
		m_bLoop = loop;
	}

	// Duration getter and setter.

	inline float CParticleDevice::Duration() const
	{
		return m_duration;
	}

	inline void CParticleDevice::Duration(float duration)
	{
		m_duration = duration;
	}

	// Playback state getter and setter.

	inline PARTICLE_PLAYBACK_STATE CParticleDevice::PlaybackState() const
	{
		return m_playbackState;
	}

	inline void CParticleDevice::PlaybackState(const PARTICLE_PLAYBACK_STATE& playbackState)
	{
		m_playbackState = playbackState;
	}

	// Particle lifetime getter and setter..
	
	inline float CParticleDevice::Lifetime() const
	{
		return m_lifetime;
	}

	inline void CParticleDevice::Lifetime(float lifetime)
	{
		m_lifetime = lifetime;
	}

	// Spawn rate getter and setter.

	inline float CParticleDevice::SpawnRate() const
	{
		return m_spawnRate;
	}

	inline void CParticleDevice::SpawnRate(float spawnRate)
	{
		m_spawnRate = spawnRate;
	}
	
	// Velocity getter and setter.
	
	inline Math::CVector3 CParticleDevice::Velocity() const
	{
		return m_velocity;
	}

	inline void CParticleDevice::Velocity(const Math::CVector3& velocity)
	{
		m_velocity = velocity;
	}

	// Initial scale getter and setter.

	inline Math::CVector3 CParticleDevice::InitialScale() const
	{
		return m_initialScale;
	}

	inline void CParticleDevice::InitialScale(const Math::CVector3& initialScale)
	{
		m_initialScale = initialScale;
	}

	// Initial color getter and setter.
	
	inline Math::CColor CParticleDevice::InitialColor() const
	{
		return m_color;
	}

	inline void CParticleDevice::InitialColor(const Math::CColor& color)
	{
		m_color = color;
	}

	// Softness getter and setter.
	
	inline float CParticleDevice::Softness() const
	{
		return m_softness;
	}

	inline void CParticleDevice::Softness(float softness)
	{
		m_softness = softness;
	}

	// Billboard type getter and setter.
	
	inline PARTICLE_BILLBOARD_TYPE CParticleDevice::BillboardType() const
	{
		return m_billboardType;
	}

	inline void CParticleDevice::BillboardType(int billboardType)
	{
		m_billboardType = (PARTICLE_BILLBOARD_TYPE)billboardType;
	}

	// Playback time getter.
	inline float CParticleDevice::PlaybackTime() const
	{
		return m_playbackTimer;
	}

	// Particle function data.
	
	inline void CParticleDevice::FunctionSetup(void* data)
	{
		m_pParticleFunctionData = data;
	}
	
	inline void CParticleDevice::FadeFunction(ParticleFade* particleFade)
	{
		m_ParticleFade = particleFade;
	}

	inline void CParticleDevice::ScaleFunction(ParticleScale* particleScale)
	{
		m_ParticleScale = particleScale;
	}

	// Max particles getter and setter.

	inline u32 CParticleDevice::MaxParticles() const
	{
		return m_instanceAllocation;
	}

	inline void CParticleDevice::MaxParticles(u32 maxParictles)
	{
		m_instanceAllocation = maxParictles;
	}

	// Instance element reference getter.
	inline CParticleDevice::PARTICLE_INSTANCE& CParticleDevice::InstanceAt(u32 index)
	{
		return *(PARTICLE_INSTANCE*)((u8*)m_pInstances + index * m_instanceStride);
	}

	// Texture filename getter and setter.

	inline std::string CParticleDevice::TextureFilename() const
	{
		return m_textureFilename;
	}

	inline void CParticleDevice::TextureFilename(const std::string& filename)
	{
		m_textureFilename = filename;
	}

	// Texture array getters and setters.

	inline CTexture** CParticleDevice::TextureArray() const
	{
		return m_ppTexture;
	}

	inline void CParticleDevice::TextureArray(CTexture** ppTexture)
	{
		m_ppTexture = ppTexture;
	}

	inline CTexture* CParticleDevice::Texture(u32 index) const
	{
		return m_ppTexture[index];
	}

	inline void CParticleDevice::Texture(CTexture* pTexture, u32 index)
	{
		m_ppTexture[index] = pTexture;
	}

	// Timer getter and setter.

	inline Util::CTimer* CParticleDevice::Timer() const
	{
		return m_pTimer;
	}

	inline void CParticleDevice::Timer(Util::CTimer* pTimer)
	{
		m_pTimer = pTimer;
	}
};

#endif
