//-------------------------------------------------------------------------------------------------
//
// Portfolio Project: Portfolio Engine
//
// File: Math\CMathMD5.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CMATHMD5_H
#define CMATHMD5_H

#include <string>
#include <sstream>
#include <iomanip>
#include "..\Globals\CGlobals.h"

namespace Math
{
	// CMD5 class.
	// Class for generating an MD5 string from a plain text string.
	// Based heavily on anothers open source class (lost the source of it awhile ago however.)
	class CMD5
	{
	public:
		CMD5();
		CMD5(const std::string& text);

		std::string							MD5(const std::string& text);
		
		void								Update(const u8 buf[], u32 length);
		void								Update(const char buf[], u32 length);

		CMD5&								Finalize();
		std::string							HexDigest() const;

		friend std::ostream&				operator <<	(std::ostream&, CMD5 md5);
		
	private:
		enum { blocksize = 64 };

		void								Initialize();

		void								Transform(const u8 block[blocksize]);
		void								Decode(u32 output[], const u8 input[], u32 length);
		void								Encode(u8 output[], const u32 input[], u32 length);
		
		bool								finalized;
		u8									buffer[blocksize];
		u32									count[2];
		u32									state[4];
		u8									digest[16];

		inline u32							F(u32 x, u32 y, u32 z);
		inline u32							G(u32 x, u32 y, u32 z);
		inline u32							H(u32 x, u32 y, u32 z);
		inline u32							I(u32 x, u32 y, u32 z);
		inline u32							RotateLeft(u32 x, int n);
		inline void							FF(u32& a, u32 b, u32 c, u32 d, u32 x, u32 s, u32 ac);
		inline void							GG(u32& a, u32 b, u32 c, u32 d, u32 x, u32 s, u32 ac);
		inline void							HH(u32& a, u32 b, u32 c, u32 d, u32 x, u32 s, u32 ac);
		inline void							II(u32& a, u32 b, u32 c, u32 d, u32 x, u32 s, u32 ac);
	};

	// F, G, H and I are basic MD5 functions.
	inline u32 CMD5::F(u32 x, u32 y, u32 z) 
	{
		return x&y | ~x&z;
	}

	inline u32 CMD5::G(u32 x, u32 y, u32 z) 
	{
		return x&z | y&~z;
	}

	inline u32 CMD5::H(u32 x, u32 y, u32 z) 
	{
		return x^y^z;
	}

	inline u32 CMD5::I(u32 x, u32 y, u32 z) 
	{
		return y ^ (x | ~z);
	}

	// rotate_left rotates x left n bits.
	inline u32 CMD5::RotateLeft(u32 x, int n) 
	{
		return (x << n) | (x >> (32-n));
	}

	// FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
	// Rotation is separate from addition to prevent recomputation.
	inline void CMD5::FF(u32 &a, u32 b, u32 c, u32 d, u32 x, u32 s, u32 ac) 
	{
		a = RotateLeft(a+ F(b,c,d) + x + ac, s) + b;
	}

	inline void CMD5::GG(u32 &a, u32 b, u32 c, u32 d, u32 x, u32 s, u32 ac) 
	{
		a = RotateLeft(a + G(b,c,d) + x + ac, s) + b;
	}

	inline void CMD5::HH(u32 &a, u32 b, u32 c, u32 d, u32 x, u32 s, u32 ac) 
	{
		a = RotateLeft(a + H(b,c,d) + x + ac, s) + b;
	}

	inline void CMD5::II(u32 &a, u32 b, u32 c, u32 d, u32 x, u32 s, u32 ac) 
	{
		a = RotateLeft(a + I(b,c,d) + x + ac, s) + b;
	}
};

#endif