//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Math\CMathTransform.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CMathTransform.h"

namespace Math
{
	CTransform::CTransform()
	{
		m_scale								= CVector3::One();
		m_localScale						= CVector3::One();

		m_pParent							= NULL;
		m_lChildren							= NULL;
	}

	CTransform::CTransform(const CTransform& other)
	{
	}

	CTransform::~CTransform()
	{
		Release();
	}

	// Method for updating a transform if it's not a child.
	void CTransform::Update()
	{
		if(!m_pParent)
		{
			UpdateMatrices();
		}
	}

	// Method for updating a transform's children.
	void CTransform::UpdateChildren()
	{
		if(m_lChildren && !m_lChildren->Empty())
		{
			Util::CListPtr::Node* node = m_lChildren->Head();
			while(node)
			{
				if(node->data)
				{
					CTransform* transform = (CTransform*)node->data;
					transform->UpdateMatrices();
				}

				node = node->next;
			}
		}
	}

	// Method updating a transforms matrices.
	void CTransform::UpdateMatrices()
	{
		D3DXMATRIX tempMtx;
		
		if(m_pParent)
		{
			m_localMtx = CMatrix4x4::Translate(m_localPosition);
			m_localMtx *= CMatrix4x4::RotateZ(m_localEuler.z);
			m_localMtx *= CMatrix4x4::RotateY(m_localEuler.y);
			m_localMtx *= CMatrix4x4::RotateX(m_localEuler.x);
			m_localMtx *= CMatrix4x4::Scale(m_localScale);
			m_localMtxInv = m_localMtx.Inverse();

			Position(m_pParent->Position() + m_pParent->WorldMatrix().TransformCoord(m_localPosition));
			Scale(m_pParent->Scale() * m_localScale);
			Euler(m_pParent->Euler() + m_localEuler);

			m_worldMtx = m_pParent->WorldMatrix() * m_localMtx;
			m_worldMtxInv = m_worldMtx.Inverse();
		}
		else
		{
			m_worldMtx = CMatrix4x4::Translate(m_position);
			m_worldMtx *= CMatrix4x4::RotateZ(m_euler.z);
			m_worldMtx *= CMatrix4x4::RotateY(m_euler.y);
			m_worldMtx *= CMatrix4x4::RotateX(m_euler.x);
			m_worldMtx *= CMatrix4x4::Scale(m_scale);
			m_worldMtxInv = m_worldMtx.Inverse();

			m_localMtx = m_worldMtx;
			m_localMtxInv = m_worldMtxInv;
		}

		UpdateChildren();
	}
	
	// Method for pointing the transform to a direction.
	void CTransform::LookAt(const CVector3& direction, const CVector3& up)
	{
		Math::CVector3 u = up;
		if(fabsf(direction.y) > Math::_1SubEpsilon)
			u = Math::CVector3(u.x, u.z, u.y);

		Math::CVector3 d = direction;
		Math::CVector3 r = (CVector3::Cross(u, d)).Normalized();
		u = (CVector3::Cross(d, r)).Normalized();

		float fwsm1 = r.x + u.y + d.z;
		float fxsm1 = r.x - u.y - d.z;
		float fysm1 = u.y - r.x - d.z;
		float fzsm1 = d.z - r.x - u.y;

		int index = 0;
		float  bfsm1 = fwsm1;
		if(fxsm1 > bfsm1)
		{
			bfsm1 = fxsm1;
			index = 1;
		}
		if(fysm1 > bfsm1)
		{
			bfsm1 = fysm1;
			index = 2;
		}
		if(fzsm1 > bfsm1)
		{
			bfsm1 = fzsm1;
			index = 3;
		}

		float bigValue = sqrtf(bfsm1 + 1.0f) * 0.5f;
		float mult = 0.25f / bigValue;

		switch(index)
		{
			case 0:
				m_rotation.w = bigValue;
				m_rotation.x = (d.y - u.z) * mult;
				m_rotation.y = (r.z - d.x) * mult;
				m_rotation.z = (u.x - r.y) * mult;
				break;
		
			case 1:
				m_rotation.x = bigValue;
				m_rotation.w = (d.y - u.z) * mult;
				m_rotation.y = (u.x + r.y) * mult;
				m_rotation.z = (r.z + d.x) * mult;
		
			case 2:
				m_rotation.y = bigValue;
				m_rotation.w = (r.z - d.x) * mult;
				m_rotation.x = (u.x + r.y) * mult;
				m_rotation.z = (d.y + u.z) * mult;
				break;
		
			case 3:
				m_rotation.z = bigValue;
				m_rotation.w = (u.x - r.y) * mult;
				m_rotation.x = (r.z + d.x) * mult;
				m_rotation.y = (d.y + u.z) * mult;
				break;
		
			default:
				break;
		}
		
		m_rotation = m_rotation.Conjugate();
		m_euler = m_rotation.ToEuler();
	}

	// Method for adding a child transform to this transform.
	void CTransform::AddChild(CTransform* child)
	{
		if(child->Parent())
		{
			child->Parent()->RemoveChild(child);
		}

		if(!m_lChildren)
			m_lChildren = new Util::CListPtr();

		child->Parent() = this;
		m_lChildren->PushBack(child);

		child->LocalPosition(m_worldMtxInv.TransformCoord(child->Position() - m_position));
		child->LocalEuler(child->Euler() - m_euler);
		child->LocalScale(child->Scale() / m_scale);
	}

	// Method for removing a child transform from this transform.
	void CTransform::RemoveChild(CTransform* child)
	{
		child->Parent() = NULL;
		m_lChildren->Remove(child);
	}

	// Method for delete a transform.
	void CTransform::Delete()
	{
		if(m_pParent)
		{
			m_pParent->RemoveChild(this);
			m_pParent = NULL;
		}

		Release();
	}

	// Method for releasing a transform.
	void CTransform::Release()
	{
		if(!m_pParent)
		{
			Release(this);
		}
		else
		{
			m_pParent->RemoveChild(this);
			Release(this);
		}
	}
	
	// Method for releasing a transform's data, and children.
	void CTransform::Release(CTransform* transform)
	{
		if(transform)
		{
			if(transform->Children())
			{
				Util::CListPtr::Node* node = transform->Children()->Head();
				while(node)
				{
					if(node->data)
					{
						// Check and release lower level children.
						Release((CTransform*)node->data);
					}

					node = node->next;
				}

				// Release this object's child list.
				SAFE_RELEASE_DELETE(transform->Children());
			}

			// No more children.
			transform->Parent() = NULL;
		}
	}
};