//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Math\CMathTransform.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CMATHTRANSFORM_H
#define CMATHTRANSFORM_H

#include <memory>
#include <d3dx9.h>
#include "CMathVector3.h"
#include "CMathMatrix4x4.h"
#include "CMathQuaternion.h"
#include "..\Utilities\CDataList.h"
#include "..\Utilities\CMemoryFree.h"

#pragma comment (lib, "d3d9.lib")
#pragma comment (lib, "d3dx9.lib")

namespace Math
{
	class CTransform;

	typedef std::shared_ptr<CTransform> CTransformSharedPtr;
	typedef std::unique_ptr<CTransform> CTransformPtr;

	// CTransform class
	// Class for creating a transformation matrix. Allows linking transforms to other transforms through children.
	class CTransform
	{
	public:
		CTransform();
		CTransform(const CTransform& other);
		virtual ~CTransform();

		void								Update();

		void								AddChild(CTransform* child);
		void								RemoveChild(CTransform* child);

		void								Delete();
		void								Release();
		
		void								LookAt(const CVector3& direction, const CVector3& up = CVector3::Up());
		
		// Directional vector inline methods.
		inline CVector3						Right();
		inline CVector3						Left();
		inline CVector3						Up();
		inline CVector3						Down();
		inline CVector3						Forward();
		inline CVector3						Back();

		// World space transformation data methods.
		inline CVector3						Position() const;
		inline void							Position(CVector3 position);

		inline CVector3						Euler() const;
		inline void							Euler(CVector3 euler);

		inline CQuaternion					Rotation() const;
		inline void							Rotation(CQuaternion rotation);
		inline void							Rotation(CVector3 euler);

		inline CVector3						Scale() const;
		inline void							Scale(CVector3 scale);

		inline CMatrix4x4					WorldMatrix() const;
		inline CMatrix4x4					WorldMatrixInv() const;
		
		// Local space transformation data methods.

		inline CVector3						LocalPosition() const;
		inline void							LocalPosition(CVector3 localPosition);

		inline CVector3						LocalEuler() const;
		inline void							LocalEuler(CVector3 localEuler);
		
		inline CQuaternion					LocalRotation() const;
		inline void							LocalRotation(CQuaternion localRotation);
		inline void							LocalRotation(CVector3 localEuler);

		inline CVector3						LocalScale() const;
		inline void							LocalScale(CVector3 localScale);

		inline CMatrix4x4					LocalMatrix() const;
		inline CMatrix4x4					LocalMatrixInv() const;

		// Miscellaneous functions.

		inline bool							HasParent() const;
		inline bool							HasChildren() const;

		inline float						DistanceFrom(Math::CVector3 point);
		inline float						DistanceFrom(CTransform* point);

	private:
		inline CTransform*&					Parent();
		inline Util::CListPtr*&				Children();

		void								UpdateChildren();
		void								UpdateMatrices();

		void								Release(CTransform* transform);

		CVector3							m_position;
		CVector3							m_localPosition;

		CVector3							m_euler;
		CVector3							m_localEuler;

		CQuaternion							m_rotation;
		CQuaternion							m_localRotation;

		CVector3							m_scale;
		CVector3							m_localScale;

		CTransform*							m_pParent;
		Util::CListPtr*						m_lChildren;

		CMatrix4x4							m_worldMtx;
		CMatrix4x4							m_worldMtxInv;
		CMatrix4x4							m_localMtx;
		CMatrix4x4							m_localMtxInv;
	};

	//
	// Directional vector inline functions.
	//

	inline CVector3 CTransform::Right()
	{
		return m_worldMtx.TransformCoord(CVector3::Right()).Normalized();
	}
	
	inline CVector3 CTransform::Left()
	{
		return m_worldMtx.TransformCoord(CVector3::Left()).Normalized();
	}
	
	inline CVector3 CTransform::Up()
	{
		return m_worldMtx.TransformCoord(CVector3::Up()).Normalized();
	}
	
	inline CVector3 CTransform::Down()
	{
		return m_worldMtx.TransformCoord(CVector3::Down()).Normalized();
	}
	
	inline CVector3 CTransform::Forward()
	{
		return m_worldMtx.TransformCoord(CVector3::Forward()).Normalized();
	}
	
	inline CVector3 CTransform::Back()
	{
		return m_worldMtx.TransformCoord(CVector3::Back()).Normalized();
	}
	
	//
	// World space transformation data functions.
	//
	
	// World position getter and setter.

	inline CVector3 CTransform::Position() const
	{
		return m_position;
	}

	inline void CTransform::Position(CVector3 position)
	{
		m_position = position;
	
		if(!m_pParent)
		{
			m_localPosition = m_position;
		}
	}

	// World euler getter and setter.

	inline CVector3 CTransform::Euler() const
	{
		return m_euler;
	}

	inline void CTransform::Euler(CVector3 euler)
	{
		m_euler = euler;
		m_rotation.RotateEuler(m_euler);

		if(!m_pParent)
		{
			m_localRotation = m_rotation;
			m_localEuler = m_euler;
		}
	}

	// World rotation getters and setters.

	inline CQuaternion CTransform::Rotation() const
	{
		return m_rotation;
	}

	inline void CTransform::Rotation(CQuaternion rotation)
	{
		m_rotation = rotation;
		m_euler = m_rotation.ToEuler();

		if(!m_pParent)
		{
			m_localRotation = m_rotation;
			m_localEuler = m_euler;
		}
	}

	inline void CTransform::Rotation(CVector3 euler)
	{
		m_rotation.RotateEuler(euler);
		m_euler = euler;

		if(!m_pParent)
		{
			m_localRotation = m_rotation;
			m_localEuler = m_euler;
		}
	}

	// World scale getter and setter.

	inline CVector3 CTransform::Scale() const
	{
		return m_scale;
	}

	inline void CTransform::Scale(CVector3 scale)
	{
		m_scale = scale;

		if(!m_pParent)
		{
			m_localScale = m_scale;
		}
	}

	// World matrix getters.

	inline CMatrix4x4 CTransform::WorldMatrix() const
	{
		return m_worldMtx;
	}

	inline CMatrix4x4 CTransform::WorldMatrixInv() const
	{
		return m_worldMtxInv;
	}

	//
	// Local space transformation data functions.
	//

	// Local position getter and setter.

	inline CVector3 CTransform::LocalPosition() const
	{
		return m_localPosition;
	}

	inline void CTransform::LocalPosition(CVector3 localPosition)
	{
		if(!m_pParent)
		{
			m_position = localPosition;
		}

		m_localPosition = localPosition;
	}

	// Local euler angle getter and setter.

	inline CVector3 CTransform::LocalEuler() const
	{
		return m_localEuler;
	}

	inline void CTransform::LocalEuler(CVector3 localEuler)
	{
		if(!m_pParent)
		{
			m_euler = localEuler;
			m_rotation.RotateEuler(m_euler);
		}

		m_localEuler = localEuler;
		m_localRotation.RotateEuler(m_localEuler);
	}

	// Local rotation getters and setters.
	
	inline CQuaternion CTransform::LocalRotation() const
	{
		return m_localRotation;
	}

	inline void CTransform::LocalRotation(CQuaternion localRotation)
	{
		if(!m_pParent)
		{
			m_rotation = localRotation;
			m_euler = m_rotation.ToEuler();
		}

		m_localRotation = localRotation;
		m_localEuler = m_localRotation.ToEuler();
	}

	inline void CTransform::LocalRotation(CVector3 localEuler)
	{
		m_localRotation.RotateEuler(localEuler);
		m_localEuler = m_localEuler;
	}

	// Local scale getter and setter.

	inline CVector3 CTransform::LocalScale() const
	{
		return m_localScale;
	}

	inline void CTransform::LocalScale(CVector3 localScale)
	{
		if(!m_pParent)
		{
			m_scale = localScale;
		}

		m_localScale = localScale;
	}

	// Local matrix getters.

	inline CMatrix4x4 CTransform::LocalMatrix() const
	{
		return m_localMtx;
	}

	inline CMatrix4x4 CTransform::LocalMatrixInv() const
	{
		return m_localMtxInv;
	}

	//
	// Child and parent data functions.
	//

	inline CTransform*& CTransform::Parent()
	{
		return m_pParent;
	}

	inline Util::CListPtr*& CTransform::Children()
	{
		return m_lChildren;
	}

	// Has parent or children getters.

	inline bool CTransform::HasParent() const
	{
		return m_pParent ? true : false;
	}

	inline bool CTransform::HasChildren() const
	{
		return m_lChildren && !m_lChildren->Empty() ? true : false;
	}

	//
	// Miscellaneous functions.
	//

	inline float CTransform::DistanceFrom(Math::CVector3 point)
	{
		return (point - m_position).Length();
	}

	inline float CTransform::DistanceFrom(CTransform* point)
	{
		return (point->Position() - m_position).Length();
	}
};

#endif