//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Math\CMathVector2.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CMathVector2.h"

namespace Math
{
	CVector2::CVector2()
	{
		x = 0.0f;
		y = 0.0f;
	}

	CVector2::CVector2(float X, float Y)
	{
		x = X;
		y = Y;
	}

	CVector2::CVector2(const CVector2& other)
	{
		x = other.x;
		y = other.y;
	}

	CVector2::CVector2(CVector2&& other)
	{
		x = other.x; other.x = 0.0f;
		y = other.y; other.y = 0.0f;
	}

	CVector2& CVector2::operator = (float f)
	{
		x = f;
		y = f;

		return *this;
	}

	CVector2& CVector2::operator = (const CVector2& v)
	{
		x = v.x;
		y = v.y;

		return *this;
	}

	CVector2 CVector2::operator - () const
	{
		return CVector2(-x, -y);
	}

	//
	// Addition operators.
	//
	
	CVector2 CVector2::operator + (const CVector2& v) const
	{
		return CVector2(x + v.x, y + v.y);
	}

	CVector2 CVector2::operator + (float f) const
	{
		return CVector2(x + f, y + f);
	}

	CVector2& CVector2::operator += (const CVector2& v)
	{
		*this = *this + v;
		return *this;
	}

	CVector2& CVector2::operator += (float f)
	{
		*this = *this + f;
		return *this;
	}

	//
	// Subtraction operators.
	//
		
	CVector2 CVector2::operator - (const CVector2& v) const
	{
		return CVector2(x - v.x, y - v.y);
	}

	CVector2 CVector2::operator - (float f) const
	{
		return CVector2(x - f, y - f);
	}

	CVector2& CVector2::operator -= (const CVector2& v)
	{
		*this = *this - v;
		return *this;
	}

	CVector2& CVector2::operator -= (float f)
	{
		*this = *this - f;
		return *this;
	}

	//
	// Multiplication operators.
	//

	CVector2 CVector2::operator * (const CVector2& v) const
	{
		return CVector2(x * v.x, y * v.y);
	}

	CVector2 CVector2::operator * (float f) const
	{
		return CVector2(x *f, y * f);
	}

	CVector2& CVector2::operator *= (const CVector2& v)
	{
		*this = *this * v;
		return *this;
	}

	CVector2& CVector2::operator *= (float f)
	{
		*this = *this * f;
		return *this;
	}

	//
	// Division operators.
	//

	CVector2 CVector2::operator / (const CVector2& v) const
	{
		return CVector2(x / v.x, y / v.y);
	}

	CVector2 CVector2::operator / (float f) const
	{
		return CVector2(x / f, y / f);
	}

	CVector2& CVector2::operator /= (const CVector2& v)
	{
		*this = *this / v;
		return *this;
	}

	CVector2& CVector2::operator /= (float f)
	{
		*this = *this / f;
		return *this;
	}

	//
	// Conditional operators.
	//

	bool CVector2::operator == (const CVector2& v) const
	{
		return (x == v.x && y == v.y);
	}

	bool CVector2::operator != (const CVector2& v) const
	{
		return (x != v.x || y != v.y);
	}
	
	//
	// Friend operators.
	//

	CVector2 operator + (float f, const CVector2& v)
	{
		return v + f;
	}

	CVector2 operator - (float f, const CVector2& v)
	{
		return v - f;
	}

	CVector2 operator * (float f, const CVector2& v)
	{
		return v * f;
	}

	CVector2 operator / (float f, const CVector2& v)
	{
		return v / f;
	}

	//
	// Vector2 static mathematic methods.
	//

	float CVector2::LengthSq(const CVector2& v)
	{
		return (v.x * v.x + v.y * v.y);
	}

	float CVector2::Length(CVector2 v)
	{
		return sqrtf(v.LengthSq());
	}

	void CVector2::Normalize(CVector2& v)
	{
		v /= Length(v);
	}

	CVector2 CVector2::Normalized(const CVector2& v)
	{
		return v / Length(v);
	}

	float CVector2::Dot(const CVector2& v0, const CVector2& v1)
	{
		return v0.x * v1.x + v0.y * v1.y;
	}

	CVector2 CVector2::Min(const CVector2& v0, const CVector2& v1)
	{
		return CVector2(min(v0.x, v1.x), min(v0.y, v1.y));
	}

	CVector2 CVector2::Max(const CVector2& v0, const CVector2& v1)
	{
		return CVector2(max(v0.x, v1.x), max(v0.y, v1.y));
	}

	//
	// Vector2 interpolation methods.
	//
	
	CVector2 MoveTowards(const CVector2& from, const CVector2& to, float t)
	{
		return CVector2(MoveTowards(from.x, to.x, t), MoveTowards(from.y, to.y, t));
	}
	
	CVector2 MoveTowards(const CVector2& from, const CVector2& to, const CVector2& t)
	{
		return CVector2(MoveTowards(from.x, to.x, t.x), MoveTowards(from.y, to.y, t.y));
	}

	CVector2 Lerp(const CVector2& from, const CVector2& to, float t)
	{
		return CVector2(Lerp(from.x, to.x, t), Lerp(from.y, to.y, t));
	}

	CVector2 Lerp(const CVector2& from, const CVector2& to, const CVector2& t)
	{
		return CVector2(Lerp(from.x, to.x, t.x), Lerp(from.y, to.y, t.y));
	}

	//
	// Vector2 variable methods.
	//

	CVector2 CVector2::Zero()
	{
		return VEC2_ZERO;
	}

	CVector2 CVector2::One() 
	{
		return VEC2_ONE;
	}

	CVector2 CVector2::Right()
	{
		return VEC2_RIGHT;
	}

	CVector2 CVector2::Left()
	{
		return VEC2_LEFT;
	}

	CVector2 CVector2::Up()
	{
		return VEC2_UP;
	}

	CVector2 CVector2::Down()
	{
		return VEC2_DOWN;
	}
};