//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Math\CMathVector3.cpp
//
//-------------------------------------------------------------------------------------------------

#include "CMathVector3.h"

namespace Math
{
	CVector3::CVector3()
	{
		x = 0.0f;
		y = 0.0f;
		z = 0.0f;
	}

	CVector3::CVector3(float X, float Y, float Z)
	{
		x = X;
		y = Y;
		z = Z;
	}

	CVector3::CVector3(const CVector3& other)
	{
		x = other.x;
		y = other.y;
		z = other.z;
	}

	CVector3::CVector3(CVector3&& other)
	{
		x = other.x; other.x = 0.0f;
		y = other.y; other.y = 0.0f;
		z = other.z; other.z = 0.0f;
	}

	CVector3& CVector3::operator = (float f)
	{
		x = f;
		y = f;
		z = f;

		return *this;
	}

	CVector3& CVector3::operator = (const CVector3& v)
	{
		x = v.x;
		y = v.y;
		z = v.z;

		return *this;
	}

	CVector3 CVector3::operator - () const
	{
		return CVector3(-x, -y, -z);
	}

	//
	// Addition operators.
	//
	
	CVector3 CVector3::operator + (const CVector3& v) const
	{
		return CVector3(x + v.x, y + v.y, z + v.z);
	}

	CVector3 CVector3::operator + (float f) const
	{
		return CVector3(x + f, y + f, z + f);
	}

	CVector3& CVector3::operator += (const CVector3& v)
	{
		*this = *this + v;
		return *this;
	}

	CVector3& CVector3::operator += (float f)
	{
		*this = *this + f;
		return *this;
	}

	//
	// Subtraction operators.
	//
		
	CVector3 CVector3::operator - (const CVector3& v) const
	{
		return CVector3(x - v.x, y - v.y, z - v.z);
	}

	CVector3 CVector3::operator - (float f) const
	{
		return CVector3(x - f, y - f, z - f);
	}

	CVector3& CVector3::operator -= (const CVector3& v)
	{
		*this = *this - v;
		return *this;
	}

	CVector3& CVector3::operator -= (float f)
	{
		*this = *this - f;
		return *this;
	}

	//
	// Multiplication operators.
	//

	CVector3 CVector3::operator * (const CVector3& v) const
	{
		return CVector3(x * v.x, y * v.y, z * v.z);
	}

	CVector3 CVector3::operator * (float f) const
	{
		return CVector3(x *f, y * f, z * f);
	}

	CVector3& CVector3::operator *= (const CVector3& v)
	{
		*this = *this * v;
		return *this;
	}

	CVector3& CVector3::operator *= (float f)
	{
		*this = *this * f;
		return *this;
	}

	//
	// Division operators.
	//

	CVector3 CVector3::operator / (const CVector3& v) const
	{
		return CVector3(x / v.x, y / v.y, z / v.z);
	}

	CVector3 CVector3::operator / (float f) const
	{
		return CVector3(x / f, y / f, z / f);
	}

	CVector3& CVector3::operator /= (const CVector3& v)
	{
		*this = *this / v;
		return *this;
	}

	CVector3& CVector3::operator /= (float f)
	{
		*this = *this / f;
		return *this;
	}

	//
	// Conditional operators.
	//

	bool CVector3::operator == (const CVector3& v) const
	{
		return (x == v.x && y == v.y && z == v.z);
	}

	bool CVector3::operator != (const CVector3& v) const
	{
		return (x != v.x || y != v.y || z != v.z);
	}
	
	//
	// Friend operators.
	//

	CVector3 operator + (float f, const CVector3& v)
	{
		return v + f;
	}

	CVector3 operator - (float f, const CVector3& v)
	{
		return v - f;
	}

	CVector3 operator * (float f, const CVector3& v)
	{
		return v * f;
	}

	CVector3 operator / (float f, const CVector3& v)
	{
		return v / f;
	}
	
	//
	// Vector3 static mathematic methods.
	//

	float CVector3::LengthSq(const CVector3& v)
	{
		return (v.x * v.x + v.y * v.y + v.z * v.z);
	}

	float CVector3::Length(CVector3 v)
	{
		return sqrtf(v.LengthSq());
	}

	void CVector3::Normalize(CVector3 v)
	{
		v /= Length(v);
	}

	CVector3 CVector3::Normalized(const CVector3& v)
	{
		return v / Length(v);
	}

	float CVector3::Dot(const CVector3& v0, const CVector3& v1)
	{
		return v0.x * v1.x + v0.y * v1.y + v0.z * v1.z;
	}
		
	CVector3 CVector3::Cross(const CVector3& v0, const CVector3& v1)
	{
		return CVector3(
			(v0.y * v1.z - v0.z * v1.y), 
			(v0.z * v1.x - v0.x * v1.z), 
			(v0.x * v1.y - v0.y * v1.x)
		); 
	}

	CVector3 CVector3::Min(const CVector3& v0, const CVector3& v1)
	{
		return CVector3(min(v0.x, v1.x), min(v0.y, v1.y), min(v0.z, v1.z));
	}

	CVector3 CVector3::Max(const CVector3& v0, const CVector3& v1)
	{
		return CVector3(max(v0.x, v1.x), max(v0.y, v1.y), max(v0.z, v1.z));
	}
	
	//
	// Vector3 interpolation methods.
	//
	
	CVector3 MoveTowards(const CVector3& from, const CVector3& to, float t)
	{
		return CVector3(MoveTowards(from.x, to.x, t), MoveTowards(from.y, to.y, t), MoveTowards(from.z, to.z, t));
	}
	
	CVector3 MoveTowards(const CVector3& from, const CVector3& to, const CVector3& t)
	{
		return CVector3(MoveTowards(from.x, to.x, t.x), MoveTowards(from.y, to.y, t.y), MoveTowards(from.z, to.z, t.z));
	}

	CVector3 Lerp(const CVector3& from, const CVector3& to, float t)
	{
		return CVector3(Lerp(from.x, to.x, t), Lerp(from.y, to.y, t), Lerp(from.z, to.z, t));
	}

	CVector3 Lerp(const CVector3& from, const CVector3& to, const CVector3& t)
	{
		return CVector3(Lerp(from.x, to.x, t.x), Lerp(from.y, to.y, t.y), Lerp(from.z, to.z, t.z));
	}

	//
	// Vector3 variable methods.
	//
	
	CVector3 CVector3::Zero()
	{
		return VEC3_ZERO;
	}

	CVector3 CVector3::One()
	{
		return VEC3_ONE;
	}

	CVector3 CVector3::Right()
	{
		return VEC3_RIGHT;
	}

	CVector3 CVector3::Left()
	{
		return VEC3_LEFT;
	}

	CVector3 CVector3::Up()
	{
		return VEC3_UP;
	}

	CVector3 CVector3::Down()
	{
		return VEC3_DOWN;
	}

	CVector3 CVector3::Forward()
	{
		return VEC3_FORWARD;
	}

	CVector3 CVector3::Back()
	{
		return VEC3_BACK;
	}
};