//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Resources\Shaders\Mesh.hlsl
//
//-------------------------------------------------------------------------------------------------

#include "RenderUtil.hlsli"

/*
	enum MESH_MAT_PARAM
	{
		MMP_ALPHA_MAP				= 0x1,
		MMP_DIFFUSE_MAP				= 0x2,
		MMP_BUMP_MAP				= 0x4,
		MMP_SPECULAR_MAP			= 0x8,
		MMP_EMISSIVE_MAP			= 0x10,
		MMP_TEXTURE					= 0x1F,
	};
*/

#ifndef SPECULAR_ON
#define SPECULAR_ON false
#endif

// Matrix constant buffer.
cbuffer MatrixBuffer : register(b0)
{
	float4x4	World;
	float4x4	WorldInv;
	float4x4	View;
	float4x4	ViewInv;
	float4x4	Projection;
	float4x4	ProjectionInv;
};

// Color constant buffer.
cbuffer ColorBuffer : register(b1)
{
	float4		AmbientColor;
	float4		DiffuseColor;
	float4		SpecularColor;
	float4		EmissiveColor;
	float		Transparency;
	int			TextureFlags;
};

// Lighting constant buffer.
cbuffer LightingBuffer : register(b2)
{
	float4		CameraPoint;
	float		SpecularFactor;
	float		Shininess;
	float		Reflectivity;
};

// Vertex input structure.
struct VIn
{
	float4 Position			: POSITION;
	float4 Normal			: NORMAL;
	float4 Binormal			: BINORMAL;
	float4 Tangent			: TANGENT;
	float4 Color			: COLOR;

	float2 TexCoord0		: TEXCOORD0;
	float2 TexCoord1		: TEXCOORD1;
	float2 TexCoord2		: TEXCOORD2;
	float2 TexCoord3		: TEXCOORD3;
};

// Vertex output structure.
struct VOut
{
	float4		Position		: SV_POSITION;
	float4		Normal			: NORMAL;
	float4		Binormal		: BINORMAL;
	float4		Tangent			: TANGENT;
	float4		Color			: COLOR;
	float4		WorldPoint		: POSITION0;
	float4		ViewPoint		: POSITION1;
	float2		TexCoord0		: TEXCOORD0;
};

// Pixel output structure.
struct POut
{
	float4		Color			: SV_TARGET0;
	float4		Diffuse			: SV_TARGET1;
	float4		Specular		: SV_TARGET2;
	float4		Normal			: SV_TARGET3;
};

// Vertex shader function.
VOut VShader(VIn input)
{
	VOut output;

	input.Position.w		=	1.0f;
	output.Position			=	mul(input.Position, World);
	output.WorldPoint		=	output.Position;
	output.Position			=	mul(output.Position, View);
	output.Position			=	mul(output.Position, Projection);
	output.ViewPoint		=	output.Position;
	
	output.Normal			=	float4(normalize(mul(float4(input.Normal.xyz, 0), World).xyz), 1.0f);
	output.Binormal			=	float4(normalize(mul(float4(input.Binormal.xyz, 0), World).xyz), 1.0f);
	output.Tangent			=	float4(normalize(mul(float4(input.Tangent.xyz, 0), World).xyz), 1.0f);

	output.Color			=	input.Color;
	output.TexCoord0		=	input.TexCoord0;

	return output;
}

Texture2D		Texture[5];

// Function for reading texture data for the texture types that are set.
void ReadTextures(inout float4 alphaMap, inout float4 diffMap, inout float4 bumpMap, 
			inout float4 specularMap, inout float4 emissiveMap, VOut input)
{
	if(TextureFlags & 0x1)
	{
		// Alpha map.
		alphaMap		= Texture[0].Sample(samplerAniso, input.TexCoord0);
	}
	if(TextureFlags & 0x2)
	{
		// Diffuse map.
		diffMap			= Texture[1].Sample(samplerAniso, input.TexCoord0);
	}
	if(TextureFlags & 0x4)
	{
		// Bump map.
		bumpMap			= Texture[2].Sample(samplerAniso, input.TexCoord0);
	}
	if(TextureFlags & 0x8)
	{
		// Speculer map.
		specularMap		= Texture[3].Sample(samplerAniso, input.TexCoord0);
	}
	if(TextureFlags & 0x10)
	{
		// Emissive map.
		emissiveMap		= Texture[4].Sample(samplerAniso, input.TexCoord0);
	}
}

// Pixel shader function.
POut PShader(VOut input) : SV_TARGET
{
	POut output = (POut)0;
	
	float4 alphaMap			= float4(1, 1, 1, 1);
	float4 diffMap			= float4(1, 1, 1, 1);
	float4 bumpMap			= float4(1, 1, 1, 1);
	float4 specularMap		= float4(1, 1, 1, 1);
	float4 emissiveMap		= float4(0, 0, 0, 1);

	float3 normal			= input.Normal.xyz;

	ReadTextures(alphaMap, diffMap, bumpMap, specularMap, emissiveMap, input);

	// Apply bumpmap to normals if one was applied.
	if(TextureFlags & 0x4)
	{
		bumpMap			= (bumpMap * 1.0f) - (0.5f);
		normal			= normalize(input.Normal.xyz + bumpMap.x * input.Tangent.xyz + bumpMap.y * input.Binormal.xyz);
	}

	// Apply the calculated color values to the four g-buffers.
	output.Diffuse = saturate(float4(diffMap.rgb * DiffuseColor.rgb, 1.0f));
	output.Specular = saturate(float4(((specularMap.rgb * SpecularColor.rgb) * specularMap.a) * SpecularFactor, Shininess / 128));
	output.Normal = saturate(float4(abs(normal.xyz), (normal.x < 0 ? 0.1f : 0.0f) + (normal.y < 0 ? 0.2f : 0.0f) + (normal.z < 0 ? 0.4f : 0.0f)));
	output.Color = saturate(float4(saturate(diffMap.rgb * float3(0.075f, 0.075f, 0.075f) + (saturate(emissiveMap.rgb + EmissiveColor.rgb) * emissiveMap.a)), alphaMap.r * Transparency));

	return output;
}