//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Resources\Shaders\Particle.hlsl
//
//-------------------------------------------------------------------------------------------------

#include "RenderUtil.hlsli"

// Matrix constant buffer.
cbuffer MatrixBuffer : register(b0)
{
	float4x4	World;
	float4x4	WorldInv;
	float4x4	View;
	float4x4	ViewInv;
	float4x4	Projection;
	float4x4	ProjectionInv;
};

// Data constant buffer.
cbuffer ConstantBuffer : register(b1)
{
	float4		Color;
	float4		CameraPosition;
	float		Softness;
	float		Time;
	float		ScreenWidth;
	float		ScreenHeight;
	int			BillboardType; // 0x0 = None, 0x1 = X, 0x2 = Y, 0x4 = Z, 0x5 = Flat, 0x7 = All
};

// Vertex input structure.
struct VIn
{
	float4	Position		: POSITION;
	float4	InstancePos		: INSTPOS;
	float4	InstanceScale	: INSTSCALE;
	float4	InstanceUV		: INSTUV;
	float4	InstanceColor	: INSTCOLOR;
	float4	InstanceData0	: INSTDATA0;
};

// Vertex output structure.
struct VOut
{
	float4	Position		: SV_POSITION;
	float4	Scale			: POSITION0;
	float4	WorldPos		: POSITION1;
	float4	Color			: COLOR;
	float4	TexCoord		: TEXCOORD0;
	float4	Depth			: TEXCOORD1;
	float4	Data0			: INSTDATA0;
};

// Pixel output structure.
struct POut
{
	float4	Color			: SV_TARGET0;
	float4	Diffuse			: SV_TARGET1;
};

bool ParticleAlive(float4 data)
{
	return (Time - data.x < data.y);
}

// Vertex shader functions.
VOut VShader(VIn input)
{
	VOut output = (VOut)0;
	
	output.Scale				=	input.InstanceScale;
	output.WorldPos				=	float4(World._14, World._24, World._34, 1.0f) + float4(input.InstancePos.xyz, 0.0f);
	
	output.Color				=	Color * input.InstanceColor;
	output.TexCoord				=	input.InstanceUV;
	output.Data0				=	input.InstanceData0;
	
	return output;
}

// Geometry shader function.
[maxvertexcount(4)]
void GShader(point VOut input[1], inout TriangleStream<VOut> outputStream)
{
	float halfSize = 0.5f;

	float3 planeNormal = (input[0].WorldPos.xyz) - CameraPosition.xyz;

	float3 upVec = float3(0, 1, 0);
	float3 rightVec = float3(1, 0, 0);
	
	if(!(BillboardType & 0x1)) // Not X.
	{
		planeNormal.x = 0;
	}
	if(!(BillboardType & 0x2)) // Not Y.
	{
		planeNormal.y = 0;
	}
	if(!(BillboardType & 0x4)) // Not Z.
	{
		planeNormal.z = 0;
	}

	planeNormal = normalize(planeNormal);
	rightVec = normalize(cross(planeNormal, upVec));
	upVec = normalize(cross(rightVec, planeNormal));

	upVec *= input[0].Scale.xyz * halfSize;
	rightVec *= input[0].Scale.xyz * halfSize;

	float3 vert[4];

	vert[0] = input[0].WorldPos.xyz - rightVec - upVec;
	vert[1] = input[0].WorldPos.xyz + rightVec - upVec;
	vert[2] = input[0].WorldPos.xyz - rightVec + upVec;
	vert[3] = input[0].WorldPos.xyz + rightVec + upVec;

	float2 texCoord[4];

	texCoord[0] = float2(0, 1);
	texCoord[1] = float2(1, 1);
	texCoord[2] = float2(0, 0);
	texCoord[3] = float2(1, 0);
	
	// Build the quad in the geometry shader.
	VOut output = input[0];
	for(int i = 0; i < 4; i++)
	{
		output.Position				=	mul(float4(vert[i], 1.0f), World);
		output.Position				=	mul(output.Position, View);
		output.Depth				=	output.Position;
		output.Position				=	mul(output.Position, Projection);

		output.TexCoord.xy			=	texCoord[i] * input[0].TexCoord.zw + input[0].TexCoord.xy;

		outputStream.Append(output);
	}

	outputStream.RestartStrip();
}

Texture2D		Texture;
Texture2D		DepthStencilTex : register(t121);

// Pixel shader function.
POut PShader(VOut input) : SV_TARGET
{
	if(!ParticleAlive(input.Data0) || input.Color.a < 0.01f)
		discard;

	float2 texCoord = 0.5f * float2(input.Position.x / ScreenWidth, -input.Position.y / ScreenHeight) + 0.5f;
	float4 depthStencil = DepthStencilTex.Load(int3(input.Position.x, input.Position.y, 0));
	
	float z = depthStencil.r;
	float x = (texCoord.x) * 2 - 1;
	float y = (1.0f - texCoord.y) * 2 - 1;
	
	float4 projectPos = float4(x, y, z, 1.0f);
	float4 viewPosition = mul(projectPos, ProjectionInv);
	viewPosition.xyz /= viewPosition.w;
	
	input.Depth.xyz /= input.Depth.w;

	// Soften the particle.
	float fade = saturate((viewPosition.z - input.Depth.z) * Softness);

	POut output;
	output.Diffuse = float4(0, 0, 0, 0);
	output.Color = Texture.Sample(samplerAnisoClamp, input.TexCoord.xy) * input.Color;
	output.Color.a *= fade;
	
	return output;
}