//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: Resources\Shaders\PostScreen.hlsl
//
//-------------------------------------------------------------------------------------------------

#include "MathUtil.hlsli"
#include "RenderUtil.hlsli"
#include "TextureProj.hlsli"
#include "PostLighting.hlsli"

// 0 = Fog Off
// 1 = Linear
// 2 = Exponential
// 3 = Exponential2

#ifndef FOG_TYPE
#define FOG_TYPE 0
#endif

// Matrix constant buffer.
cbuffer MatrixBuffer : register(b0)
{
	float4x4	World;
	float4x4	WorldInv;
	float4x4	View;
	float4x4	ViewInv;
	float4x4	Projection;
	float4x4	ProjectionInv;
};

// Data constant buffer.
cbuffer ConstantBuffer : register(b1)
{
	float4		Scale;
	float3		CameraPoint;
	int			PassIndex;
};

cbuffer FogBuffer : register(b2)
{
	float4		FogColor;
	float2		FogExtents;
	float		FogDensity;
};

// Vertex input structure.
struct VIn
{
	float4 Position			: POSITION;
	float2 TexCoord0		: TEXCOORD;
};

// Vertex output structure.
struct VOut
{
	float4 Position			: SV_POSITION;
	float2 TexCoord0		: TEXCOORD0;
	float3 ViewPoint		: POSITION0;
};

/*
	Texture (t0): Ambient
	Texture (t1): Diffuse
	Texture (t2): Specular
	Texture (t3): Normal
	Texture (t4): Depth Stencil
*/

Texture2DArray	gBufferTextures		: register(t120);
Texture2D		DepthStencilTex		: register(t121);

// Vertex shader function.
VOut VShader(VIn input)
{
	VOut output = (VOut)0;

	input.Position.w	= 1.0f;
	input.Position.xyz	*= Scale.xyz;
	output.Position		= mul(input.Position, World);
	output.Position		= mul(output.Position, View);
	output.ViewPoint	= output.Position.xyz;
	output.Position		= mul(output.Position, Projection);

	output.TexCoord0	= input.TexCoord0;

	return output;
}

// Function for adding for to the world.
float Fog(float depth)
{
	#if FOG_TYPE == 1
		// Linear fog
		return (FogExtents.y - depth) / (FogExtents.y - FogExtents.x);
	#elif FOG_TYPE == 2
		// Exponential fog
		return 1.0f / pow(_ln, depth * FogDensity);
	#elif FOG_TYPE == 3
		// Exponential2 fog
		return 1.0f / pow(pow(_ln, depth * FogDensity), 2);
	#endif

	return 1.0f;
}

// Pixel shader function.
float4 PShader(VOut input) : SV_TARGET
{
	float4 output = float4(0, 0, 0, 1);
	
	float4 ambient			= saturate(gBufferTextures.Sample(samplerAnisoClamp, float3(input.TexCoord0.xy, 0)));
	float4 diffuse			= saturate(gBufferTextures.Sample(samplerAnisoClamp, float3(input.TexCoord0.xy, 1)));
	float4 specular			= saturate(gBufferTextures.Sample(samplerAnisoClamp, float3(input.TexCoord0.xy, 2)));
	float4 normalColor		= saturate(gBufferTextures.Sample(samplerAnisoClamp, float3(input.TexCoord0.xy, 3)));
	float4 depthStencil		= saturate(DepthStencilTex.Sample(samplerAnisoClamp, input.TexCoord0.xy));

	float4 normal = normalColor;

	int normW = round(normal.w * 10.0f);
	normal.x *= normW & 1 ? -1.0f : 1.0f;
	normal.y *= normW & 2 ? -1.0f : 1.0f;
	normal.z *= normW & 4 ? -1.0f : 1.0f;

	float4 color = ambient;
	float4 projectionColor = float4(0, 0, 0, 0);

	// Projection, view, and world positions from depth stencil render target.
	float z = depthStencil.r;
	float x = (input.TexCoord0.x) * 2 - 1;
	float y = (1.0f - input.TexCoord0.y) * 2 - 1;
	
	float4 projectPos = float4(x, y, z, 1.0f);
	float4 viewPosition = mul(projectPos, ProjectionInv);
	viewPosition /= viewPosition.w;
	float4 worldPosition = mul(viewPosition, ViewInv);
	worldPosition.xyz /= worldPosition.w;
	
	if(diffuse.a > 0.5f)
	{
		float4 lightColor = float4(0, 0, 0, 1);
		int i;

		// Texture projection.
		#if TEXTUREPROJ_COUNT
		TextureProjector texProj = (TextureProjector)0;

		for(i = 0; i < TEXTUREPROJ_COUNT; i++)
		{
			texProj = TEXTUREPROJ[i];

			float4 projColor = ProjectTexture(color.rgb, worldPosition.xyz, texProj);
			projectionColor = saturate(float4((projectionColor.rgb * (1.0f - projColor.a)) + 
				(projColor.rgb * (projColor.a)), projectionColor.a + projColor.a));
		}

		#endif
		
		//
		// Lighting.
		//
		
		// Directional lights.
		#if DIRLIGHT_COUNT
		DirectionalLight dirLight = (DirectionalLight)0;
		
		for(i = 0; i < DIRLIGHT_COUNT; i++)
		{
			dirLight = DIR_LIGHT[i];

			DirectionalLighting(diffuse.rgb, specular, normal.xyz, CameraPoint.xyz, worldPosition.xyz, 
				dirLight, i, lightColor);
		}
		#endif

		// Point lights.
		#if POINTLIGHT_COUNT
		PointLight pointLight = (PointLight)0;

		int j;

		for(i = 0; i < POINTLIGHT_COUNT; i++)
		{
			pointLight = POINT_LIGHT[i];

			PointLighting(diffuse.rgb, specular, normal.xyz, CameraPoint.xyz, worldPosition.xyz, 
				pointLight, i, lightColor);
		}
		#endif

		// Spotlights.
		#if SPOTLIGHT_COUNT
		Spotlight spotlight = (Spotlight)0;
		
		for(i = 0; i < SPOTLIGHT_COUNT; i++)
		{
			spotlight = SPOTLIGHT[i];

			SpotLighting(diffuse.rgb, specular, normal.xyz, CameraPoint.xyz, worldPosition.xyz, 
				spotlight, i, lightColor);
		}
		#endif
		
		color.rgb = saturate(color.rgb + lightColor.rgb);
	}

	// Fog.
	if(diffuse.a > 0.25f && FOG_TYPE != 0)
	{
		float fogFactor		= Fog(viewPosition.z);
		color.rgb			= fogFactor * color.rgb + (1.0f - fogFactor) * FogColor.rgb;
	}
	
	output.rgb = (saturate(color.rgb) * (1.0f - projectionColor.a)) + (projectionColor.rgb * projectionColor.a);

	return output;
}