//-------------------------------------------------------------------------------------------------
//
// Copyright (c) 2015-2016 Ryan Alasandro
//
// Portfolio Project: Portfolio Engine
//
// File: UI\CUIFontRenderer.h
//
//-------------------------------------------------------------------------------------------------

#ifndef CUIFONTRENDERER_H
#define CUIFONTRENDERER_H

#include "..\Graphics\CGeoObject.h"
#include "CUIFreeType.h"

namespace UI
{
	// Enumerators for horizontal text alignment.
	enum class UI_HALIGN
	{
		Left,
		Center,
		Right,
	};
	
	// Enumerators for vertical text alignment.
	enum class UI_VALIGN
	{
		Top,
		Center,
		Bottom,
	};

	const float FontScaler = 0.01f;
	const float FontSDFScale = 6.66666667f;

	// CFontRenderer class
	// Class for creating a new font renderer.
	class CFontRenderer : public GFX::CGeoObject
	{
	public:
		// Free type font vertex structure.
		struct FT2_VERTEX
		{
			Math::CVector3					position;

			FT2_VERTEX()
			{
			}

			FT2_VERTEX(Math::CVector3 Position)
			{
				position = Position;
			}
		};

		// Free type font instance structure.
		struct FT2_INSTANCE
		{
			Math::CVector3					position;
			Math::CVector3					scale;
			Math::CVector4					uvOffset;
			Math::CColor					color;
		};

	public:
		CFontRenderer();
		CFontRenderer(const CFontRenderer& other);
		virtual ~CFontRenderer();

		int									Initialize() override;
		int									Update() override;
		int									Render() override;
		void								Release() override;

		int									Print(const std::string& text, const Math::CVector3& position, float size, const Math::CColor& color, float fade, const UI_HALIGN& hAlign, const UI_VALIGN& vAlign);
		int									Print(const std::string& text, const Math::CVector3& position, float size, const Math::CColor& color, float fade, const UI_HALIGN& hAlign, const UI_VALIGN& vAlign, Math::CVector4* targetBounds);
		
		float								CalculateLineLength(const std::string& text, float size);
		int									ClosestIndexRelativeToPoint(const Math::CVector2& point, const std::string& text, const Math::CVector3 position, float size, const UI_HALIGN& hAlign, const UI_VALIGN& vAlign, Math::CVector4* targetBounds);

		inline Math::CVector4				LastBounds();
		inline float						LastLineLength();

		inline CFreeType*					FreeType();
		inline void							FreeType(CFreeType* pFreeType);

	private:
		Math::CVector2						CalculateStartingPoint(const std::string& text, const Math::CVector3& position, const Math::CVector2& scale, const UI_HALIGN& hAlign, const UI_VALIGN& vAlign, Math::CVector4* targetBounds);
		
		float								GetKernValue(char c0, char c1);

		inline Math::CVector3				CalculateScaleFromSize(float size);
		inline float						CalculateSizeFromScale(const Math::CVector3& scale);

		float								m_lineLength;
		Math::CVector4						m_lastBounds;

		float								m_kerning;

		CFreeType*							m_pFreeType;
	};

	// Last font bounds getter. (For use with intersections)
	inline Math::CVector4 CFontRenderer::LastBounds()
	{
		return m_lastBounds;
	}

	// Last line length getter.
	inline float CFontRenderer::LastLineLength()
	{
		return m_lineLength;
	}

	// FreeType2 object getter and setter.

	inline CFreeType* CFontRenderer::FreeType()
	{
		return m_pFreeType;
	}

	inline void CFontRenderer::FreeType(CFreeType* pFreeType)
	{
		m_pFreeType = pFreeType;
	}
	
	// Method for calculating the font and SDF scale from a font size.
	inline Math::CVector3 CFontRenderer::CalculateScaleFromSize(float size)
	{
		Math::CVector2 scale = Math::CVector2::One() * size * FontScaler;
		float z = scale.x * FontSDFScale; // The 'Z' component is used for sdf calculations.
		return Math::CVector3(scale.x * m_pFreeType->m_ptScale, scale.y * m_pFreeType->m_ptScale, z);
	}

	// Method for calculating the font size from the font scale.
	inline float CFontRenderer::CalculateSizeFromScale(const Math::CVector3& scale)
	{
		return  (scale.x / m_pFreeType->m_ptScale / FontScaler);
	}
};

#endif